;-----------------------------------------------------------------------------
; NexGen Nx586 BIOS Hypercode
;-----------------------------------------------------------------------------
; File: CACHE.ASM - 
; Date: 2004-08-15
;
; Disassembled by Dipl.-Inf. Herbert Oppmann (herby@memotech.franken.de)
;-----------------------------------------------------------------------------

		include COMMON.INC

	public	loc_4C20
	public	loc_4C44

; see US patent 5,905,997
; "Set-associative cache memory utilizing a single bank of physical memory"

; first level cache
; =================
; Code: 16 KB, 4-Way, 32 Byte/Line, Dual-ported, SI, LRU
; Data: 16 KB, 4-Way, 32 Byte/Line, Dual-ported, MESI, LRU
;
; In addition, where the L2 cache is a 4-way set associative cache, an indi-
; cation of which of four banks the data is stored in is included with the
; status information stored in the L1 cache. This eliminates the need for a
; tag look-up in the L2 cache for writes which hit in the L1 cache since the
; data is known to be present in the L2 cache if there is a hit in the L1
; cache, and the only information needed is which bank the data is going to
; be stored in.
; In addition to the conventional status information such as valid and dirty,
; the L1 tags store "type" bits to classify a cache block. These bits, for
; example, may indicate that the cache block is defective so that a block of
; memory is never allocated to it, thus permitting the use of a microprocessor
; with some number of defects in its on-chip cache. Certain status information
; for the L2 cache is also stored with the status information of the L1 cache.
; This enhances the speed of operations by avoiding the cycle for tag look-up
; in the L2 cache during a write operation.

; +--------------------------------+-------------------+--------------------+
; | tag address                    | V3 | V2 | V1 | V0 | S1 | S0 | B1 | B 0 |
; +--------------------------------+-------------------+--------------------+
;                                   <-   L1 status   -> <- L2 status in L1 ->
; Vx = Valid subblock
; S1,S0 = MESI state
; B1,B0 = bank
;
; access to first level cache in hypermode:
; direct read-write:
; code or data:
; bank 0: 40000000h-40000FFFh
; bank 1: 40400000h-40400FFFh
; bank 2: 40800000h-40800FFFh
; bank 3: 40C00000h-40C00FFFh
; data or code:
; bank 0: 50000000h-50000FFFh
; bank 1: 50400000h-50400FFFh
; bank 2: 50800000h-50800FFFh
; bank 3: 50C00000h-50C00FFFh
;
; This is probably the tag area. It is read-write (DWORD size) on the first
; address of each cache line, i.e. every 20h bytes.
; code or data:
; bank 0: 42000000h-420000FE0h
; bank 1: 42400000h-424000FE0h
; bank 2: 42800000h-428000FE0h
; bank 3: 42C00000h-42C000FE0h
; data or code:
; bank 0: 52000000h-520000FE0h
; bank 1: 52400000h-524000FE0h
; bank 2: 52800000h-528000FE0h
; bank 3: 52C00000h-52C000FE0h
;
; US Patent 5,511,175 (for the pre-Nx586 9-chip implementation):
; The instruction cache tag RAM contains the address tag, a "Valid" bit,
; and an "Attention" bit for each line in the ICache. The Attention bit
; indicates that the DEC chip may also have data from this line cached
; in the BPC. 
; The Attention bit of a cache line is set when a copy of one or more
; instruction words resident within that line are entered into a BPC entry.
; Then, when such a cache line is invalidated or otherwise modified, the set
; Attention bit signals the need to do a BPC consistency check. While any
; associated instruction copies may have already been replaced from the BPC,
; any copies still resident are detected and immediately invalidated.

; The data cache tag RAM contains the address tag and line state bits for
; each line in the DCache. The possible line states are Absent, Shared Read,
; Owned Clean, and Owned Dirty, supporting a write back multiprocessor cache
; coherency protocol (modified write once).


; second level cache
; ==================
;
; Unified: 256 KB or 1 MB, 4-Way, 32 Byte/Line, Single-ported, MESI
; SRAMs Contain Tags and Microcode Update
; US Patent 5,745,724:
; .. configurable cache sizes for up to 2 MB of data and tag storage ..
;
; organisation (256 KB):
; 8 chips of 32Kx8, resulting in 32K words of 64 bit
; cache = 4 banks (numbered 0,1,2,3) of 64 KB
; each bank consists of 2048 blocks (lines) of 32 byte
; each block consists of 4 subblocks (words) of 64 bits (8 byte)
;
;     <------- 64 bits word -------->
;
;       0   1   2   3   4   5   6   7   <- chip no.
; 0K +---+---+---+---+---+---+---+---+
;    |                               | ^
;    |                               | |
;    |                               | |
;    |                               | bank 0
;    |                               | |
;    |                               | |
;    |                               | v
; 8K +---+---+---+---+---+---+---+---+
;    |                               | ^
;    |                               | |
;    |                               | |
;    |                               | bank 1
;    |                               | |
;    |                               | |
;    |                               | v
;16K +---+---+---+---+---+---+---+---+
;    |                               | ^
;    |                               | |
;    |                               | |
;    |                               | bank 2
;    |                               | |
;    |                               | |
;    |                               | v
;24K +---+---+---+---+---+---+---+---+
;    |                               | ^
;    |                               | |
;    |                               | |
;    |                               | bank 3
;    |                               | |
;30K +---+---+---+---+---+---+---+---+ |
;    | address tags and status bits  | v 512 blocks
;32K +---+---+---+---+---+---+---+---+
;
;
; tag area in more detail:
;
;       16
;     <-bit->
;30K +-------+-------+-------+-------+
;    |       |       |       |       | ^
;    |       |       |       |       | |
;    |       |       |       |       | |
;.25 |       |       |       |       | |
;    |       |       |       |       | |
;    |       |       |       |       | |
;    |       |       |       |       | |
;.5  |       |       |       |       | |
;    |       |       |       |       | |
;    |  for  |  for  |  for  |  for  | |
;    |       |       |       |       |
;.75 |       |       |       |       | 4-way associative
;    |       |       |       |       |
;    | bank0 | bank1 | bank2 | bank3 | |
;    |       |       |       |       | |
;31K |       |       |       |       | |
;    |       |       |       |       | |
;    |       |       |       |       | |
;    |       |       |       |       | |
;.25 |       |       |       |       | |
;    |       |       |       |       | |
;    |       |       |       |       | |
;    |       |       |       |       | |
;.5  |       |       |       |       | |
;    |       |       |       |       | |
;    |       |       |       |       | |
;    |       |       |       |       | |
;.75 |       |       |       |       | |
;    |       |       |       |       | v
;    |       |       |       +-------+
;    |       |       |       |unused | 3-way associative
;32K +-------+-------+-------+-------+
;
;
; tag in detail:
;
;   15   14   13   12                                                      0
; +----+----+----+-----------------------------------------------------------+
; | S1 | S0 |A29*|                      A28-A16                              |
; +----+----+----+-----------------------------------------------------------+
; S1,S0 MESI state: 11=modified, 10=exclusive, 01=shared, 00=invalid
; A29* = A29 for normal accesses and 1 for hypermode accesses
; A31-30 is 0 for normal accesses and 1 for hypermode accesses
;
; Access             SRAM address lines
;               C14  C13  C12  C11  C10-C2  C1  C0
; Tag            1    1    1    1   A13-A5  A15 A14
; Data Bank 3    1    1   A15  A14  A13-A5  A4  A3
; Data Bank 2    1    0   A15  A14  A13-A5  A4  A3
; Data Bank 1    0    1   A15  A14  A13-A5  A4  A3
; Data Bank 0    0    0   A15  A14  A13-A5  A4  A3
;
; C14-C0 = address lines to the cache SRAMs
; A15-A3 = address from the L2 controller
;
; The present invention also provides local read-write storage for use by
; the processor by allocating a number of L2 cache blocks. The data in these
; cache blocks is not accessible to normal programs, and there is no need
; for physical main memory which corresponds to these locations. This area
; would be useful for certain reserved ("hypercode") storage for system
; management software or writable control programs for the microprocessor.
; By specifications, hypercode accesses may be cached only in bank 3 if
; A31=A30=A29=1. Memory addresses may be cached in banks 3, 2, 1 or 0 if
; A31=A30=A29=0. Memory accesses with A31=A30=0 and A29=1 may be cached only
; in logical banks 2, 1 or 0. Neither memory nor hypercode accesses with
; A15=A14=1 may be cached in logical bank 3 since it is the area reserved for
; tags. Note that the degree of associativity in the Level 2 cache changes
; with the type of address space (degree of 1 for hypercode and degree of 3 or
; 4 for memory) and the address range (degree of 3 or 4 for memory, depending
; on the address range). Certain address ranges are not cacheable. By speci-
; fication, hypercode cache blocks are never replaced by memory cache blocks
; in Level 2, giving hypercode priority over memory for cache replacements.
; Even though the hypercode is direct mapped (degree of 1), it is effectively
; "locked" into the Level 2 cache.
; The logical bank 3, not occupied by the tags, may be used to cache either
; hypercode or memory address space. However, priority is given to hypercode
; in cache replacements. Normally, the least recently used (LRU) cache line
; is replaced if needed to make space for a new entry. However, if bank 3
; block is occupied by hypercode, it will not participate in the LRU algo-
; rithms to make space for a new block of physical memory. A hypercode
; entry may be replaced by another hypercode entry.
;
; access to second level cache in hypermode:
; direct read-write:
; bank 0: 60000000h-6000FFFFh for 256KB, -6003FFFFh for 1MB
; bank 1: 60400000h-6040FFFFh for 256KB, -6043FFFFh for 1MB
; bank 2: 60800000h-6080FFFFh for 256KB, -6083FFFFh for 1MB
; bank 3: 60C00000h-60C0FFFFh for 256KB, -60C3FFFFh for 1MB
; for 256KB cache size: each bank is mirrored four times
; see HYPRES.ASM, second level cache bank 3 test
;
; hypercode (in bank 3):
; 60C30000h-60C3DFFFh	48 KB
; (on 256KB cache size this is mirrored to 60C00000h-60C0DFFFh)
;
; tag area (in bank 3):
; 60C0C000h-60C0FFFFh
; 60C1C000h-60C1FFFFh	\
; 60C2C000h-60C2FFFFh	 | 1 MB cache only
; 60C3C000h-60C3FFFFh	/
;
; a byte read on each cache line (every 32 byte) flushes(??) the cache line:
; bank 0: 62000000h-6200FFE0h for 256KB, -6203FFE0h for 1MB
; bank 1: 62400000h-6240FFE0h for 256KB, -6243FFE0h for 1MB
; bank 2: 62800000h-6280FFE0h for 256KB, -6283FFE0h for 1MB
; bank 3: 62C00000h-62C0FFE0h for 256KB, -62C3FFE0h for 1MB

; TBD: what about LRU? Where is the info for this?
; On a 4-way set-associative cache, LRU would need to remember 4! = 24 sequences,
; needing 5 bits per cache line. Probably a simpler pseudo-LRU scheme is used
; which only needs 3 bits per cache line: one bit to select between two groups
; of 2 entries, and two bits for selecting the entry in the respective group.

 IF V033S OR V033V
		db '@(#)cache.asm',TAB,'3.7',0
 ELSEIF V033W
		db '@(#)cache.asm',TAB,'3.8',0
 ELSE
		db '@(#)cache   ',0
 ENDIF
		ALIGN0	32

; Invalidate 1st and 2nd level cache
loc_4C20:	@CALL	loc_4C68	; invalidate 1st level cache
		@CALL	GetCacheSize	; EBX = #of DWORDs
		@CALL	loc_4D22	; invalidate 2nd level cache
		@RET

; Invalidate 1st level cache, flush(??) the 2nd level cache
loc_4C44:	@CALL	loc_4C68	; invalidate 1st level cache
		@CALL	GetCacheSize	; EBX = #of DWORDs
		@CALL	loc_4DC1	; flush(??) 2nd level cache
		@RET

; invalidate first level cache (code and data)
; The first level cache must always be invalidated when manipulating
; the second level cache because each first level cache entry also
; keeps the 2nd level cache MESI-bits and bank number.
; The 1st level data cache does not need to be flushed because of the
; write-through to the 2nd level cache. All modified cache lines in the
; 1st level cache are also modified in the 2nd level cache!
; code: 16 KB, 4-Way, 32 Byte/Line, Dual-ported, SI, LRU
; data: 16 KB, 4-Way, 32 Byte/Line, Dual-ported, MESI, LRU
loc_4C68:	mov	edx, h_ccfg[ebp]
		and	edx, 0FDFFDFFFh	; clear bit 25 and 13
		@CALL	WriteCacheConf
	; clear (fill with 0) dwords at:
	; 42000000h, 42000020h, 42000040h, ... 420000FE0h
	; 42400000h, 42400020h, 42400040h, ... 424000FE0h
	; 42800000h, 42800020h, 42800040h, ... 428000FE0h
	; 42C00000h, 42C00020h, 42C00040h, ... 42C000FE0h
		xor	eax, eax
		xor	edi, edi
		mov	ebx, 42C00000h
		mov	ecx, 42800000h
		mov	edx, 42400000h
		mov	esi, 42000000h
loc_4C93:
 IF VVL
		bt	dword ptr [ebx], 30
		jc	short loc_509B
		mov	[ebx], eax
loc_509B:	bt	dword ptr [ecx], 30
		jc	short loc_50A3
		mov	[ecx], eax
loc_50A3:	bt	dword ptr [edx], 30
		jc	short loc_50AB
		mov	[edx], eax
loc_50AB:	bt	dword ptr [esi], 30
		jc	short loc_50B3
		mov	[esi], eax
loc_50B3:	clc	
 ELSE
		mov	[ebx], eax
		mov	[ecx], eax
		mov	[edx], eax
		mov	[esi], eax
 ENDIF
		inc	edi
		add	ebx, CLINESIZE
		add	ecx, CLINESIZE
		add	edx, CLINESIZE
		add	esi, CLINESIZE
		cmp	edi, 80h
		jnz	short loc_4C93
		mov	edx, h_ccfg[ebp]	; restore bit 25 and 13
		@CALL	WriteCacheConf
	; clear (fill with 0) dwords at:
	; 52000000h, 52000020h, 52000040h, ... 520000FE0h
	; 52400000h, 52400020h, 52400040h, ... 524000FE0h
	; 52800000h, 52800020h, 52800040h, ... 528000FE0h
	; 52C00000h, 52C00020h, 52C00040h, ... 52C000FE0h
		xor	eax, eax
		xor	edi, edi
		mov	ebx, 52C00000h
		mov	ecx, 52800000h
		mov	edx, 52400000h
		mov	esi, 52000000h
loc_4CD5:
 IF VVL
		bt	dword ptr [ebx], 30
		jc	short loc_50F6
		mov	[ebx], eax
loc_50F6:	bt	dword ptr [ecx], 30
		jc	short loc_50FE
		mov	[ecx], eax
loc_50FE:	bt	dword ptr [edx], 30
		jc	short loc_5106
		mov	[edx], eax
loc_5106:	bt	dword ptr [esi], 30
		jc	short loc_510E
		mov	[esi], eax
loc_510E:	clc
 ELSE
		mov	[ebx], eax
		mov	[ecx], eax
		mov	[edx], eax
		mov	[esi], eax
 ENDIF
		inc	edi
		add	ebx, CLINESIZE
		add	ecx, CLINESIZE
		add	edx, CLINESIZE
		add	esi, CLINESIZE
		cmp	edi, 80h
		jnz	short loc_4CD5
		@RET

; subroutine
; write to cache configuration register at address 10000000h
; E: EDX
; V: EAX
; R:-
WriteCacheConf:	mov	eax, 10000000h
		mov	[eax], edx
		DELAYIO	3
		@RET


; subroutine
; E:-
; R: EBX = L2 cache size (as number of dwords)
GetCacheSize:	bt	h_dbflag[ebp], 19	; cache size?
		mov	ebx, 40000h		; 1 MB
		jc	short GetCSizeRet
		mov	ebx, 10000h		; 256 KB
GetCSizeRet:	@RET


; subroutine
; invalidate second level cache (but don't disturb hypercode in bank 3!)
; Note: this works on the tags!
; E: EBX = L2 cache size (as number of dwords)
loc_4D22:	; do bank 0, 1 and 2
		xor	eax, eax		; state invalid
		mov	edi, 60C0C000h		; bank 3, tags
		mov	ecx, ebx
		shr	ecx, CLINESHIFT		; size of tag area
loc_4D2E:	mov	[edi], eax		; invalidate tag for bank 0+1
		mov	[edi+4], ax		; inv. tag for bank 2
		add	edi, 8
		or	edi, 0C000h
		loop	loc_4D2E
		; do bank 3
		@CALL	loc_4D9A
		; EBX is still L2 cache size
		; ECX is end address of hypercode
		; EDX is start address of hypercode
		mov	edi, 0			; offset into bank 3
loc_4D4E:	cmp	edi, edx
		jl	short loc_4D58		; before hypercode
		cmp	edi, ecx
		jg	short loc_4D58		; after hypercode
		; in hypercode: don't disturb!
		jmp	short loc_4D6D

loc_4D58:	@CALL	loc_4D7A
		or	esi, 60C0C006h		; base address of tags for bank 3
		mov	word ptr [esi],	0	; invalidate tag for bank 3
loc_4D6D:	add	edi, CLINESIZE
		cmp	edi, ebx		; end of 2nd level cache?
		jl	short loc_4D4E
		@RET

; subroutine
; calculate offset into array to corresponding offset into tag area
; E: EDI = offset into second level cache, bank 3 (into array)
; V: EAX
; R: ESI = offset into second level cache, bank 3 (into tag area)
loc_4D7A:	mov	esi, edi
		and	esi, 0C000h		; bit A15-A14
		shr	esi, 11			; into bit 4-3 (C1-C0)
		mov	eax, edi
		and	eax, 3FFE0h		; A17-A5
		or	esi, eax		; -> bit 17-5 (C14-C2)
		or	esi, 0C000h		; set bit 15-14 (C12-11)
						; (into tag area)
		@RET

; subroutine
; where is hypercode in second level cache bank 3?
; E:-
; V: EAX
; R: ECX = end address of hypercode (as offset from beginning of bank 3)
;    EDX = start address of hypercode (as offset from beginning of bank 3)
loc_4D9A:	bt	h_dbflag[ebp], 19	; cache size?
		mov	eax, 40000h-CLINESIZE	; 1M
		jc	short loc_4DAB
		mov	eax, 10000h-CLINESIZE	; 256K
loc_4DAB:	mov	edx, h_base[ebp]
		mov	ecx, edx
		and	edx, eax
		add	ecx, h_endhyp[ebp]	; end of hypercode
		dec	ecx
		and	ecx, eax
		@RET

; subroutine
; flush(??) second level cache (but don't disturb hypercode in bank 3!)
; E: EBX = L2 cache size (as number of dwords)
loc_4DC1:	; do bank 0, 1 and 2
		mov	eax, 62800000h
		mov	esi, 62400000h
		mov	edx, 62000000h
		mov	ecx, ebx
		shr	ecx, CLINESHIFT
loc_4DD5:	test	byte ptr [eax],	0
		test	byte ptr [esi],	0
		test	byte ptr [edx],	0
		add	eax, CLINESIZE
		add	esi, CLINESIZE
		add	edx, CLINESIZE
		loop	loc_4DD5
		; do bank 3
		@CALL	loc_4D9A
		; EBX is still L2 cache size
		; ECX is end address of hypercode
		; EDX is start address of hypercode
		mov	edi, 0			; offset into bank 3
loc_4DF8:	cmp	edi, edx
		jl	short loc_4E02		; before hypercode
		cmp	edi, ecx
		jg	short loc_4E02		; after hypercode
		; in hypercode: don't disturb!
		jmp	short loc_4E0D

loc_4E02:	mov	esi, edi
		or	esi, 62C00000h		; base address of bank 3
		test	byte ptr [esi],	0
loc_4E0D:	add	edi, CLINESIZE
		cmp	edi, ebx		; end of 2nd level cache?
		jl	short loc_4DF8
		@RET

		ALIGN	256
hypseg		ends
		end

;--------------------------------------------------------------- end CACHE.ASM