		%noincl
;-----------------------------------------------------------------------------
; NexGen Nx586 BIOS Hypercode
;-----------------------------------------------------------------------------
; File: COMMON.INC - macros, defines
; Date: 2020-09-24
;
; Disassembled by Dipl.-Inf. Herbert Oppmann (herby@memotech.franken.de)
;-----------------------------------------------------------------------------

		.386p
		title "NexGen Nx586 BIOS Hypercode"
		%list
		%macs

; versions: (keep in sync with makefile)
 IFIDNI %VERS,<'033s'>
VVL	= 's'
VPCI	= 0
 ELSEIFIDNI %VERS,<'033v'>
VVL	= 'v'
VPCI	= 0
 ELSEIFIDNI %VERS,<'033w'>
VVL	= 'w'
VPCI	= 0
 ELSEIFIDNI %VERS,<'034i'>
VVL	= 0
VPCI	= 'i'
 ELSEIFIDNI %VERS,<'034n'>
VVL	= 0
VPCI	= 'n'
 ELSEIFIDNI %VERS,<'034p'>
VVL	= 0
VPCI	= 'p'
 ELSEIFIDNI %VERS,<'034q'>
VVL	= 0
VPCI	= 'q'
 ELSE
ERR "VERS incorrect"
 ENDIF

V033S	= VVL  EQ 's'
V033V	= VVL  EQ 'v'
V033W	= VVL  EQ 'w'
V034I	= VPCI EQ 'i'
V034N	= VPCI EQ 'n'
V034P	= VPCI EQ 'p'
V034Q	= VPCI EQ 'q'

hypseg		segment public 'CODE' use32
		assume cs:hypseg, ds:hypseg
		assume ss:nothing, es:nothing, fs:nothing, gs:nothing

; Exception vectors
EXCP_DEBUG	equ	1	;#DB
EXCP_NMI	equ	2	; Non-maskable external interrupt
EXCP_INVALIDOP	equ	6	;#UD
EXCP_NOMATH	equ	7	;#NM
EXCP_DOUBLEFAULT equ	8	;#DF
EXCP_INVALIDTSS	equ	10	;#TS
EXCP_SEGNOTPRES	equ	11	;#NP
EXCP_STACKFAULT	equ	12	;#SS
EXCP_GENPROT	equ	13	;#GP
EXCP_PAGEFAULT	equ	14	;#PF

; memory addresses
 IF VVL
M_BIOSBASE	equ	0FFFE4000h
 ELSE	; PCI
M_BIOSBASE	equ	0FFFE0000h
 ENDIF

; port addresses
P_CMOSADR	equ	70h
P_CMOSVAL	equ	71h
P_DIAG		equ	80h

; ASCII codes
TAB		equ	9

; Cache
CLINESIZE	equ	32	; size of a cache line (4 x 64 bit)
CLINESHIFT	equ	5

; bit masks
BIT		equ	<1 shl >

; jump delay
DELAYJMP	macro	n:=<1>
		rept n
		 jmp	short $+2
		endm
		endm

; nop delay
DELAYNOP	macro	n:=<1>
		rept n*8
		 nop
		endm
		endm

; I/O delay
DELAYIO 	macro	n:=<1>
		rept n
		 in	al, P_DIAG
		 endm
		endm

; output value to manufacturer's diagnostic port
DIAG		macro	n:req
		mov	al,n
		out	P_DIAG, al
		endm

; subroutine call via push/jmp combination
; Uses db 0E9h to force 32-bit near jmp on local destinations.
; "extrn" destinations use 0E9h anyway and can't be used with
; dd dst-m: linker calculates extrn label minus distance of "m"
; from start of this module. The base address of the current
; module is not considered. What a crap!
@CALL		macro	dst:req
		local	m
		push	offset m
	if 80h and ( .type dst )
		jmp	dst
	else
		db	0E9h
		dd	dst-m
	endif
m:		endm

; jump
@JMP		macro	dst:req
		add	dst, h_base[ebp]
		jmp	dst
		endm

; return from subroutine called by @CALL
@RET		macro
		pop	eax
		@JMP	eax
		endm


; Hypercode code-macros
; =====================

; 32-bit general registers
r32_eax		equ 0
r32_ecx		equ 1
r32_edx		equ 2
r32_ebx		equ 3
r32_esp		equ 4
r32_ebp		equ 5
r32_esi		equ 6
r32_edi		equ 7

;_________________________
; Prefix
; Byte     BANK=0  BANK=1
;_________________________
; 0x26     ES      HS
; 0x2E     CS      CS
; 0x36     SS      SS
; 0x3E     DS      IS
; 0x64     FS      TR
; 0x65     GS      IS
;_________________________

; normal segments
rsg_es		equ 0
rsg_cs		equ 1
rsg_ss		equ 2
rsg_ds		equ 3
rsg_fs		equ 4
rsg_gs		equ 5
rsg_?6		equ 6	; old SS
rsg_?7		equ 7	; old CS

; other segments
rosg_ldtr	equ 0
rosg_tr		equ 1
rosg_gdtr	equ 2
rosg_idtr	equ 3
rosg_hbr	equ 4	; hypercode base register
; 5-7 not used


; The internal Descriptor Cache Registers consist of
; - 32 bit AR (Access rights / Limit)
; - 32 bit linear base address

; AR:
; bit definitions for A/R
bar_p		equ 31
bar_dpl		equ 29 ;-30
bar_s		equ 28
; SYSTEM (bar_s = 0)
;  Type equ 24-27
;  0000	reserved
;  0001	16-Bit TSS (available)
;  0010	LDT
;  0011	16-Bit TSS (busy)
;  0100	16-Bit Call Gate
;  0101	Task Gate
;  0110	16-Bit Interrupt Gate
;  0111	16-Bit Trap Gate
;  1000	Reserved
;  1001	32-Bit TSS (available)
;  1010	Reserved
;  1011	32-Bit TSS (busy)
;  1100	32-Bit Call Gate
;  1101	Reserved
;  1110	32-Bit Interrupt Gate
;  1111	32-Bit Trap Gate
; APPLICATION (bar_s = 1)
bar_cd		equ 27
; DATA (bar_cd =0)
;Expand-down     equ 26
;Write           equ 25
;Accessed        equ 24
; CODE (bar_cd =1)
;Conforming      equ 26
;Readable        equ 25
;Accessed        equ 24

;Granularity     equ 23
;D/B             equ 22
;RES             equ 21      ; \ MIGHT BE RESERVED
;AVL             equ 20      ; /  OR NOT USED
;LIM             equ 00-19

; Table 1 - A/R and Limit
; -----------------------
; R/M[31]         P	present: 1=present
; R/M[30:29]      DPL	descriptor privilege level
; R/M[28]         S	descriptor type: 0=system, 1=application
; R/M[27:24]      TYPE	segment type
; R/M[23]         G	granularity: 0=byte, 1=4KByte
; R/M[22]         D/B	if code segment: default operand size (0=16 bit, 1=32 bit)
;			if data segment: big (max. offset size 0=0x0000FFFF, 1=0xFFFFFFFF)
; R/M[21]         RES	reserved (=0)
; R/M[20]         AVL	available
; R/M[19:0]       Limit[19:0]


; 0F,E0		RETH - return from hypermode to mortal mode
RETH		macro
		db 0Fh,0E0h
		endm

; 0F,E1,IMM8	INTH - return from hypermode to mortal mode
;		and immediately generate exception n
INTH		macro	n:req
		db 0Fh,0E1h,n
		endm

; 0F,E2		??, see EXCPH.ASM
OP_0FE2		macro
		db 0Fh,0E2h
		endm

; 0F,E3		not used

; 0F,E4		??
OP_0FE4		macro
		db 0Fh,0E4h
		endm

; 0F,E5		not used
; 0F,E6		not used
; 0F,E7		not used
; 0F,E8		not used
; 0F,E9		not used
; 0F,EA		not used

; 0F,EB /0	WRCFG0 Ev - write configuration register 0
;		(this register can't be read back, so a copy of the
;		value written is kept in h_cfg0 by hypercode)
WRCFG0		macro r:req
		db 0Fh,0EBh,0C0h+r32_&r
		endm

; 0F,EB /1	OP_0FEB_1 Ev - ???? not used

; 0F,EB /2	WRCFG1 Ev - write configuration register 1
;		(this register can't be read back, so a copy of the
;		value written is kept in h_cfg1 by hypercode)
WRCFG1		macro r:req
		db 0Fh,0EBh,0D0h+r32_&r
		endm

; 0F,EB /3	WRSBNK Ev - write segment bank flag
;		for segment override prefixes, only bit 0 relevant
WRSBNK		macro r:req
		db 0Fh,0EBh,0D8h+r32_&r
		endm

; 0F,EB /4	WRCPL Ev - write CPL
WRCPL		macro r:req
		db 0Fh,0EBh,0E0h+r32_&r
		endm

; 0F,EB /5	RDCPL Ev - read CPL
RDCPL		macro r:req
		db 0Fh,0EBh,0E8h+r32_&r
		endm

; 0F,EB /6	OP_0FEB_6 Ev - read CPU model + stepping
;		bit 31:28  not model
;		bit 27:24  not stepping
;		bit 23:0   unused
RDMOST		macro r:req
		db 0Fh,0EBh,0F0h+r32_&r
		endm

; 0F,EB /7	OP_0FEB_7 Ev - ???? not used

; 0F,F0;0F,F1	WRB Sw,Ev - write segment linear base address
WRB		macro s:req, r:req
		ifdef rsg_&s
		  db 0Fh,0F0h,0C0h+(rsg_&s shl 3)+r32_&r ;; normal segment
		else
		  db 0Fh,0F1h,0C0h+(rosg_&s shl 3)+r32_&r ;; other segment
		endif
		endm

; 0F,F2;0F,F3	RDB Ev,Sw - read segment linear base address
RDB		macro r:req, s:req
		ifdef rsg_&s
		  db 0Fh,0F2h,0C0h+(rsg_&s shl 3)+r32_&r ;; normal segment
		else
		  db 0Fh,0F3h,0C0h+(rosg_&s shl 3)+r32_&r ;; other segment
		endif
		endm

; 0F,F4;0F,F5	WRAR Sw,Ev - write segment A/R and limit
WRAR		macro s:req, r:req
		ifdef rsg_&s
		  db 0Fh,0F4h,0C0h+(rsg_&s shl 3)+r32_&r ;; normal segment
		else
		  db 0Fh,0F5h,0C0h+(rosg_&s shl 3)+r32_&r ;; other segment
		endif
		endm

; 0F,F6;0F,F7	RDAR Ev,Sw - read segment A/R and limit
RDAR		macro r:req, s:req
		ifdef rsg_&s
		  db 0Fh,0F6h,0C0h+(rsg_&s shl 3)+r32_&r ;; normal segment
		else
		  db 0Fh,0F7h,0C0h+(rosg_&s shl 3)+r32_&r ;; other segment
		endif
		endm

; 0F,F8;0F,F9	WRSEL Sw,Ev - write segment selector
WRSEL		macro s:req, r:req
		ifdef rsg_&s
		  db 0Fh,0F8h,0C0h+(rsg_&s shl 3)+r32_&r ;; normal segment
		else
		  db 0Fh,0F9h,0C0h+(rosg_&s shl 3)+r32_&r ;; other segment
		endif
		endm

; 0F,FA		LOPCMSK r,Ib - write to Opcode Mask Register
; write IMM8 to 16 byte (64 x 2 bit) Opcode Mask Register
LOPCMSK		macro r:req, n:req
		db 0Fh,0FAh,0C0h+r,n
		endm

; To ensure backward compatibility the encoding r=0-7 programs the least
; significant bit of the mask register and r=8-F programs the most sig-
; nificant bit. Though the opcode mask register is 64 x 2 bits, only
; locations 60-0 are allowed to be loaded, with locations 63-61 being
; hard-wired as follows:
;______________________________________
; bit  value           reason
;______________________________________
; 63   01              To cover the #UD opcodes (always DECEXCP)
; 62   0, (HFBIT||HyperEn)
;                      To allow hypermode instructions
; 61   01 if in hypermode,
;                      To disallow HLT in hypermode
;______________________________________
; 00 - Normal operation
; 01 - DECEXCP p-op issued resulting in hypermode entry
; 10 - Force issue of instruction to AP
; 11 - Hold issue of instruction until no p-op's are outstanding

;   Grp#      BIOS    Macro Instruction
; dec  hex   byte:bit Groupings
; 00   00     0:0     ***unused***
; 01   01     0:1     MOV reg/mem to/from reg/mem
; 02   02     0:2     MOV imm to reg/mem
; 03   03     0:3     MOV reg/mem to/from segreg
; 04   04     0:4     MOVSX, MOVZX
; 05   05     0:5     PUSH reg/mem/imm
; 06   06     0:6     PUSH segreg
; 07   07     0:7     POP reg/mem (except ESP)
; 08   08     1:0     POP segreg
; 09   09     1:1     PUSHA, POPA
; 10   0a     1:2     XCHG
; 11   0b     1:3     IN
; 12   0c     1:4     OUT
; 13   0d     1:5     LEA
; 14   0e     1:6     LDS, LES, LFS, LGS, LSS
; 15   0f     1:7     CLC, CLD, CLI, CLTS, CMC
; 16   10     2:0     LAHF, SAHF
; 17   11     2:1     POPF
; 18   12     2:2     PUSHF
; 19   13     2:3     STC, STD, STI
; 20   14     2:4     ADD, ADC, INC, DEC, SUB, SBB, CMP, NEG
; 21   15     2:5     AAA, AAS, AAD, AAM, DAA, DAS
; 22   16     2:6     MUL
; 23   17     2:7     IMUL
; 24   18     3:0     DIV
; 25   19     3:1     IDIV
; 26   1a     3:2     ROL, ROR, SAL, SAR, SHL, SHR
; 27   1b     3:3     RCL, RCR
; 28   1c     3:4     SHLD, SHRD, CBW, CWD
; 29   1d     3:5     AND, TEST, OR, XOR, NOT
; 30   1e     3:6     CMPS, SCAS
; 31   1f     3:7     INS
; 32   20     4:0     LODS, STOS
; 33   21     4:1     MOVS
; 34   22     4:2     OUTS
; 35   23     4:3     XLAT
; 36   24     4:4     BSF, BSR, BT, BTC, BTR, BTS
; 37   25     4:5     CALL near
; 38   26     4:6     CALL far
; 39   27     4:7     JMP near
; 40   28     5:0     JMP far
; 41   29     5:1     RET near
; 42   2a     5:2     RET far
; 43   2b     5:3     Jcc
; 44   2c     5:4     LOOP, LOOPZ/E, LOOPNZ/NE
; 45   2d     5:5     SETcc
; 46   2e     5:6     ENTER, LEAVE
; 47   2f     5:7     INT 3, INT n
; 48   30     6:0     INTO
; 49   31     6:1     BOUND
; 50   32     6:2     IRET
; 51   33     6:3     POP ESP
; 52   34     6:4     MOV CR, DR, TR
; 53   35     6:5     NOP, WAIT
; 54   36     6:6     Prefixes (0F, 66, 67, 2E, 3E, 26, 36, 64, 65, F0, F2, F3)
; 55   37     6:7     ARPL
; 56   38     7:0     LAR, LSL
; 57   39     7:1     LGDT, LIDT, LLDT, SGDT, SIDT, SLDT, LTR, STR
; 58   3a     7:2     LMSW, SMSW
; 59   3b     7:3     VERR, VERW
; 60   3c     7:4     NP instructions
; 61   3d     7:5     HLT
; 62   3e     7:6     Hypermode instructions
; 63   3f     7:7     #UD


; Use "MOV Ev,Sw" to read normal segment selectors ...
; NOTE: Segments #6 and #7 can also be used (including in MOV SReg instructions).

; 0F,FB		RDSEL Ev,Sw - read segment selector
RDSEL		macro r:req, s:req
		db 0Fh,0FBh,0C0h+(rosg_&s shl 3)+r32_&r
		endm

; MOV Sw,Ev and MOV Ev,Sw also with CS, ?6 and ?7
MOVSEL		macro r1:req, r2:req
		ifdef rsg_&r1
		  db 8Eh,0C0h+(rsg_&r1 shl 3)+r32_&r2	;; MOV Sw,Ev
		else
		  db 8Ch,0C0h+(rsg_&r2 shl 3)+r32_&r1	;; MOV Ev,Sw
		endif
		endm


; Floating Point Instructions
; Code created by the ESC pseudo-instruction:
; ESC 00ee emmm -> 1101 1eee ..mm m... with second byte < 0C0h

; D9 /1 = Read FPU configuration dword, a copy of which is kept in h_fcfg
FRDCFG		macro l:req
		ESC 09h,l
		endm

; D9,EF ?? Used in FPTAN.ASM and FSINCOS.ASM
FNUNKNOWN1	macro
		db 0D9h,0EFh
		endm
; Maybe it reduces st to the [0,pi/2) range typical of trigonometry, and
; sets flags according to the result.

; DB /4 Load FPU Tag Word
FNLDTW		macro l:req
		ESC 1Ch,l
		endm

; DB /6 Store FPU Tag Word
FNSTTW		macro l:req
		ESC 1Eh,l
		endm

; DB,E4 287 instruction
FNSETPM		macro
		db 0DBh,0E4h
		endm

; DD /1 ?? Writes a word to memory, used in NPHCODE.ASM
FNUNKNOWN2	macro l:req
		ESC 29h,l
		endm

; DD /5 Load FPU Status Word (opposite of official FNSTSW)
FNLDSW		macro l:req
		ESC 2Dh,l
		endm

; DF /1 write FPU configuration dword, a copy of which is kept in h_fcfg
FWRCFG		macro l:req
		ESC 39h,l
		endm

h_ftemp		equ	-7Ah	; temporary extended real
; FPU operating environment and register stack for hypercode,
; accessed via EBP register by negative offset
; (emulation always does "fnsave byte ptr h_fenviron[ebp]" first)
h_fenviron	equ	-70h
h_fcw		equ	-70h	; ----	Control Word
h_fsw		equ	-6Ch	; ----	Status Word
h_ftw		equ	-68h	; ----	Tag Word
h_fipoffs	equ	-64h	; IP Offset
h_fopcodecs	equ	-60h	; Opcode CS Selector
h_fdataoffs	equ	-5Ch	; Data Operand Offset
h_fopsec	equ	-58h	; ----	Operand Selector
h_fr0		equ	-54h	; R0
h_fr1		equ	-4Ah	; R1
; -040h		R2
; -036h		R3
; -02Ch		R4
; -022h		R5
; -018h		R6
; -00Eh		R7
; -004h		saved CR0 (see loc_8E80 and loc_8F53)

; hypercode variables, accessed via EBP register by positive offset
hypvar		struc
	;+000h
h_uart		dd	?	; UART port to use (0= none, 2F8h or 3F8h)
h_excph		dd	?	; procedure to be called if an exception occurs
h_hbr		dd	?	; copy of hyper base register (points to hyper
				; mode entry points and hypermode structure)
h_base		dd	?	; base address of hypercode (in 2nd level cache)
	;+010h
h_dbflag	dd	?	; debug flags
		; 31:25	unused
		; 24	??
		; 23:20	unused
		; 19	cache size: 0=256K 1=1M
		; 18	unused
		; 17	if set, we should generate #DB when exiting
		; 16	??
		; 15:3	unused
		;  2	set if DR7 was saved in h_dr7 (usually for temporarily
		;	disabling breakpoint traps when accessing code)
		;  1	set if we should set EFLAGS.RF=1
		;  0	set if #DB generated by single-stepping (i.e.,
		;	we should set DR6.BS=1)
h_ucs		dd	?	; user CS?? (used in RETF.ASM)
		dd	?	; ?? (used in FSCALE.ASM)
		dd	?	; unused
	;+020h
h_content	dd	?	; register or memory contents (byte, word or dword)
h_disp		dd	?	; memory displacement saved here (if memory operand)
h_sib		dd	?	; SIB byte
h_modrm		dd	?	; modR/M byte
	;+030h
h_opcode	dw	?	; word opcode; if 0Fxxh high byte is 0Fh
h_segpre	db	?	; latest segment override prefix: 00h = none,
				; 26h = ES, 2Eh = CS, 36h = SS
				; 3Eh = DS, 64h = FS, 65h = GS
		db	?	; unused
h_opflag	dd	?
		; 31	set if 32-bit address size, clear if 16-bit
		; 30	set if 32-bit operand size, clear if 16-bit
		; 29	set if LOCK prefix found
		; 28	set if REP prefix found
		; 27	set if REPNZ prefix found
		; 26:24	segment override: 000 = none, 001 = ES, 010 = CS,
		;	011 = SS, 100 = DS, 101 = FS, 110 = GS
		; 23	set if 0Fh opcode
		; 22	unused
		; 21	set if operand is byte
		; 20	if set on entry of subroutine GetOperand:
		;	Only calculate the effective address from modR/M and SIB,
		;	but do not fetch the value from there into h_content
		; 19	set if address size prefix (67h) found
		; 18	set if operand size prefix (66h) found
		; 17:8	unused
		;  7:0	length of opcode
		dd	?	; unused in PCI, used in VL HYPRES.ASM
h_ccfg		dd	?	; copy of mem 10000000h (cache configuration register)
				; see also h_ccfgmaster
		; init = 84901000h (VL) + bit 28 cache size
		; 1098 7654 3210 9876 5432 1098 7654 3210
		; 100x.0100.1001.0000.0001.0000.0000.0000
		; init = 8CD01300h (PCI) + bit 28 cache size
		; 1098 7654 3210 9876 5432 1098 7654 3210
		; 100x.1100.1101.0000.0001.0011.0000.0000
		; 28	second level cache 0=256K, 1=1M
		; 26	see INVOP.ASM
		; 25	see CACHE.ASM
		; 18	level 2 cache: 0=off, 1=on
		; 14	see INVOP.ASM
		; 13	see CACHE.ASM
		; Bits 25 and 13 are cleared temporarily while working on
		; first level cache tags at 42000000h range.
		; Bit 26 and 25 perform two functions on the same data set
		; and bit 14 and 13 perform the same functions on another
		; data set. Maybe 25 and 13 are for "use the cache" and
		; 26 and 14 are for "update the cache" i.e. switch on the
		; replacement strategy (works only if bit 25/13 is on).
; US Patent 5,745,724:
; .. configurable cache sizes for up to 2 MB of data and tag storage ..
; Maybe bit 29 is for 2nd level cache size, too? Or is that for the AMD K6?
; US Patent 5,745,732:
; The level-2 cache can be software-configured for write-through rather than
; write-back operation.
	;+040h
h_cfg0		dd	?	; copy of configuration register 0 contents
		; init = 5DF974C4h
		; 1098 7654 3210 9876 5432 1098 7654 3210
		; 0101.1101.1111.1001.0111.0100.1100.0100
		; 31	used to mask the DECEXCP p-op
		; 24	level 1 cache: 0=off, 1=on
h_cfg1		dd	?	; copy of configuration register 1 contents
		; init = 1507144Fh (V033S), 15071C4Fh (V033V OR V033W) or
		; 37021C0Fh (PCI)
		; 23	set by MOV SS,CS:xx when game assistant is on (VL only)
		;	see DecMOVSEG in EMUL.ASM, set also in ForceExec in
		;       DECODE.ASM
		; 20	set/reset by setting UART for HDEB in INVOP.ASM
		; 18    reset by INTO.ASM, RETF.ASM, TASK.ASM and XESC.ASM,
		;	set in INVOP.ASM

; US Patent 5,454,117:
; Configuration Control includes 3 components:
; (1) PDMaskSel<2..0>,
; (2) ConfigNoBPC, and
; (3) ConfigLoopT.
; When ConfigNoBPC is active, the Hit/Miss information is ignored, and the
; branch is treated as though there was always a hit in the BPC.
; Activating ConfigLoopT causes the special case instructions that are condi-
; tional but are taken much more often than not (e.g. LOOP) to be always
; predicted taken.
; US Patent 5,768,575:
; Although the late arriving Page Frame Address is not a problem for addressing
; the cache, the full 32 bit physical address is needed early in order to
; complete the write pipeline address comparisons. There are two possible ways
; for the read queue pipeline to solve this problem. MCS implements both
; solutions, and hypercode sets a configuration bit to select the mode. In one
; mode, only the Offset portion of PADR is compared against the write queue
; pipeline when the read queue pipeline is processing a read directly from PADR.
; If there is a match on the Page Offset, the read is stalled, and tried again
; with the full address on the following cycle. If there is no match on the
; page offset, the read is allowed to proceed. In the other mode, the PREV_R0
; register (described in the Padr pipeline section) is used to predict the PADR
; Page Frame address. Once the actual Page Frame Address arrives, it compares
; itself to the predicted version and stalls the read if there is a mismatch.
; Since subsequent reads and writes are often to the same page of memory, the
; PREV_R0 register, which saves the last valid write page address, is a good
; predictor of the actual PADR page address.

; setting a bit causes all debug related `hits` to be trapped to Hyper Code
; setting a bit (XESC) causes all non-debug exceptions and faults to be trapped
;   to Hyper Code
; setting a bit causes all `shutdown` events to be trapped to Hyper Code
; setting a bit causes all external interrupts to be trapped to Hyper Code
h_opcmsk0	dd	?	; copy of opcode mask register (see LOPCMSK)
		; 18	\ normally 0, set to 1 for ID on (EFLAGS.AC toggle)
		; 17	/ causes emulation of POPF and PUSHF
h_opcmsk1	dd	?	; copy of opcode mask register (see LOPCMSK)
	;+050h
h_inibase	dd	?	; base address of CS, SS, DS, FS, GS,
				; LDTR, TR, GDTR and IDTR after reset
				; (= constant 0FFFF0000h)
h_inisel	dd	?	; selector value of CS, SS, DS, FS, GS,
				; LDTR, TR, ?6, ?7 and base address of ES
				; after reset (= constant 0000F000h)
h_inieip	dd	?	; EIP after reset (= constant 0000FFF0h)
		dd	?	; used (FPU)
	;+060h
		dd	?	;  "
		dd	?	;  "
		dd	?	;  "
		; These four values are only used in HYPRES.ASM. Read back
		; after writing h_fcfg, h_fcfg+40h, h_fcfg+80h and h_fcfg+0C0h
h_fcfg		dd	?	; copy of floating point configuration register
		; init = 00800010h
	;+070h
h_npx		dd	?	; 81h if FPU present on-chip, else 0DEADBEEFh
h_tsst		dd	?
		; 31	set if TSS T bit is set
		; 30 \	set in TASK.ASM when hypermode entry via offset 0C00h
		; 29 /	from h_gatflag.30-29 when not clear
		; 28	set in TASK.ASM when hypermode entry via offset 0C00h
		;	and h_gatflag.30-29 are clear
		; 27	set in TASK.ASM when hypermode entry via offset 0D00h
h_tsssel	dd	?	; new TSS selector
h_trsel		dd	?	; TR.sel
	;+080h
		dd	?	; unused
h_eip1		dd	?	; EIP #1
		dd	?
		dd	?
	;+090h
		dd	4 dup(?)
	;+0A0h
		dd	?
h_endhyp	dd	?	; end of (uncompressed) hypercode
		dd	?
		dd	?
	;+0B0h
h_dbstate	dd	88 dup(?) ; CPU state (B0-207, 158h byte) for serial debugging
	;+210h
		dd	?	; h_bufsiz, size in byte of buffer pointed to by h_buffer
		dd	?	; h_buffer, buffer (linear address)
		dd	?
		dd	?	; h_tmpesp, ESP is temporarily saved here
	;+220h
		dd	?	; h_cr0
; 4 x 2 DWORDS, using EDX as index
; 224 EBX0, 22C EBX1, 234 EBX2, 23C EBX3
; 228 ECX0, 230 ECX1, 238 ECX2, 240 ECX3
; is memory copy of values written to:
; 1E000000h, 1E400000h, 1E800000h, 1EC00000h
; 1C000000h, 1C400000h, 1C800000h, 1CC00000h
		dd	?
		dd	?
		dd	?
	;+230h
		dd	?
		dd	?
		dd	?
		dd	?
	;+240h
		dd	?	; used by INVOP.ASM
		dd	3 dup(?)
	;+250h
		dd	?	; copy of mem 12000000h
		dd	?	; copy of mem 12400000h
		dd	?	; copy of mem 12800000h
		dd	?	; copy of mem 12C00000h
	;+260h
		dd	?	; copy of mem 16000000h
		dd	?	; copy of mem 16400000h
		dd	?	; copy of mem 14000000h
		dd	?	; copy of mem 14400000h
	;+270h
		dd	?	; copy of mem 16800000h
		dd	?	; copy of mem 16C00000h
		dd	?	; copy of mem 14800000h
		dd	?	; copy of mem 14C00000h
; US Patent 5,768,575:
; The Page based cacheability registers in MCS are loaded by hypercode, and
; can be programmed to allow certain areas of the physical page frame address
; to be specified as MMIO, write through, or write protected.
	;+280h
h_ccfgmaster	dd	?	; master for h_ccfg
		; init = 8ADF7000h (VL)
		; 1098 7654 3210 9876 5432 1098 7654 3210
		; 1000.1010.1101.1111.0111.0000.0000.0000
		; init = 8ADF7300h (PCI)
		; 1098 7654 3210 9876 5432 1098 7654 3210
		; 1000.1010.1101.1111.0111.0011.0000.0000
h_opcmsk2	dd	?	; copy of opcode mask register (see LOPCMSK)
h_opcmsk3	dd	?	;  "
		dd	?	; copy of mem 1D000000h
	;+290h
		dd	?	; copy of mem 1A000000h
		dd	?	; copy of mem 18000000h

; Floating point statistics: count number of FPU instructions handled by hypercode
h_f2xm1c	dd	?	; count FSIN instructions
h_fyl2x		dd	?	; count FYL2X instructions
	;+2A0h
h_fptanc	dd	?	; count FPTAN instructions
h_fpatanc	dd	?	; count FPATAN instructions
h_fxtractc	dd	?	; count FXTRACT instructions
h_fyl2xp1	dd	?	; count FYL2XP1 instructions
	;+2B0h
h_fsqrtc	dd	?	; count FSQRT instructions
h_fsincosc	dd	?	; count FSINCOS instructions
h_fscalec	dd	?	; count FSCALE instructions
h_fsinc		dd	?	; count F2XM1 instructions
	;+2C0h
h_fcosc		dd	?	; count FCOS instructions
h_fbldc		dd	?	; count FBLD instructions
h_fbstpc	dd	?	; count FBSTP instructions

h_mode		dd	?
		; 31:16	bit 31 set in TASK.ASM, upper 16 bit used in
		;	INVOP.ASM as counter/error flag in cache check
		; 15:8	unused
		;  7	486 instructions enable
		;  6:0	unused
	;+2D0h
		dd	22 dup(?)	; code loc_16D0
h_cr2		dd	?	; CR2
h_dr7		dd	?	; DR7
	;+330h
		dd	?	; used in RETF.ASM and INT.ASM in V033W
		dd	?	; copy of segment register bank (VL only)
		dd	?	; used in INT.ASM and INVOP.ASM
		dd	?	; unused
	;+340h
h_eflsh		dd	?	; EFLAGS shadow
		; 31:19	unused?
		; 18	AC
		; 17:0	unused?
		dd	?	; used in FBSTP.ASM, IO_EM.ASM and NPHCODE.ASM
				; for saving copy of CR0
		dd	?	; unused
		dd	?	;  "
	;+350h
		dd	31 dup (?) ; reserved for stack
h_edi		dd	?	; EDI
	;+3D0h
h_esi		dd	?	; ESI
h_edx		dd	?	; EDX
h_ecx		dd	?	; ECX
h_ebx		dd	?	; EBX
	;+3E0h
h_eax		dd	?	; EAX
h_ebp		dd	?	; EBP
;-- stack is filled by CPU on hypercode entry up to this point
; on entry 0800h (exception)
; h_gate	31-16	exception code
;		15-0	selector
; h_gatflag

; on entry 0880h (soft reset)
; on entry 0900h (#UD undefined opcode)

; on entry 0A00h (JMP and CALL)
; h_gate
; h_gatflag	30
; ?6 = 
; ?7 = destination segment
; used while CALL emulation
; h_gate
; h_gatflag	31	16/32-bit call gate
;		30	from entry
;		20:16	parameter count from call gate

; on entry 0A80h (far return)
; on entry 0B00h (interrupt)
; on entry 0B80h (interrupt return)
; on entry 0C00h (task switch)
; on entry 0C80h (I/O bit map)
; on entry 0D00h (?task)
; on entry 0D80h (INTO instruction)
; on entry 0F00h (FPU)


h_gate		dd	?	; call gate (or other gate) offset
h_gatflag	dd	?
		; 31	set to 1 if 32-bit call gate or maybe use ESP
		;	when you push whatever needs to be pushed
		; 30	? used in CALL.ASM
		; 29	?
		; 28	?
		; 27	set if we should NOT clear IF (else set IF=0)
		;NOTE: 28:27 is also copied to ?7[1:0]
		; 26	trap bit; if clear, fault
		; 25	?
		; 24	set if error code
		; 23:16	Exception/interrupt nr
		; 15:0	Exception nr	used in RETF.ASM, IRET.ASM and DECODE.ASM
		; 7, 6: see INVOP.ASM
	;+3F0h
h_eip3		dd	?	; next sequential PC
h_efl		dd	?	; EFLAGS
h_esp		dd	?	; ESP
h_eip2		dd	?	; current PC
	;+400h
hypvar		ends

ModBase:	; for alignment

ALIGN0		macro n:req
		db n-(($-ModBase) AND (n-1)) dup(0)
		endm

;-------------------------------------------------------------- end COMMON.INC