;-----------------------------------------------------------------------------
; NexGen Nx586 BIOS Hypercode
;-----------------------------------------------------------------------------
; File: EMSUBS.ASM - 
; Date: 2020-09-27
;
; Disassembled by Dipl.-Inf. Herbert Oppmann (herby@memotech.franken.de)
;-----------------------------------------------------------------------------

		include COMMON.INC

	extrn	FORCEBIG:abs
	extrn	GetOperand:near
	extrn	loc_2AA7:near

	public	GetOperandAndReg
	public	WriteRM
	public	WriteReg
	public	WriteMem
	public	CheckCPL0

		db '@(#)emsubs.asm',TAB,'3.2',0
		ALIGN0	32

; ===========================================================================
;		S u b r	o u t i	n e
;
; Read both R/M and register operands
; V: EAX, ..
; R: EBX=reg field
;    ECX=reg value (word values are sign-extended to 32 bits, byte values are not!)
GetOperandAndReg: @CALL	GetOperand	; parse ModR/M (+SIB+displacement) bytes, read r/m operand
		; read register operand
		movzx	ecx, byte ptr h_modrm[ebp]
		and	ecx, 38h	; reg field
		shr	ecx, 3
		mov	ebx, ecx	; save reg field
		shl	ecx, 1		; *2
		add	ecx, offset loc_7043
		@JMP	ecx

loc_7043:	jmp	short loc_7056	; AL/AX/EAX
		jmp	short loc_7061	; CL/CX/ECX
		jmp	short loc_706C	; DL/DX/EDX
		jmp	short loc_7077	; BL/BX/EBX
		jmp	short loc_7082	; AH/SP/ESP
		jmp	short loc_70A2	; CH/BP/EBP
		jmp	short loc_70C2	; DH/SI/ESI
		jmp	loc_70E2	; BH/DI/EDI

loc_7056:	mov	ecx, h_eax[ebp]
		jmp	loc_7100

loc_7061:	mov	ecx, h_ecx[ebp]
		jmp	loc_7100

loc_706C:	mov	ecx, h_edx[ebp]
		jmp	loc_7100

loc_7077:	mov	ecx, h_ebx[ebp]
		jmp	loc_7100

loc_7082:	test	byte ptr h_opflag+2[ebp], 20h	; byte operand?
		jnz	short loc_7090		; AH
		mov	ecx, h_esp[ebp]
		jmp	short loc_7100

loc_7090:	mov	cl, byte ptr h_eax+1[ebp]
		and	ecx, 0FFh
		@RET

loc_70A2:	test	byte ptr h_opflag+2[ebp], 20h	; byte operand?
		jnz	short loc_70B0		; CH
		mov	ecx, h_ebp[ebp]
		jmp	short loc_7100

loc_70B0:	mov	cl, byte ptr h_ecx+1[ebp]
		and	ecx, 0FFh
		@RET

loc_70C2:	test	byte ptr h_opflag+2[ebp], 20h	; byte operand?
		jnz	short loc_70D0		; DH
		mov	ecx, h_esi[ebp]
		jmp	short loc_7100

loc_70D0:	mov	cl, byte ptr h_edx+1[ebp]
		and	ecx, 0FFh
		@RET

loc_70E2:	test	byte ptr h_opflag+2[ebp], 20h	; byte operand?
		jz	short loc_70FA		; BH
		mov	cl, byte ptr h_ebx+1[ebp]
		and	ecx, 0FFh
		@RET

loc_70FA:	mov	ecx, h_edi[ebp]
loc_7100:	; register value is in ECX
		test	byte ptr h_opflag+2[ebp], 20h	; byte operand?
		jz	short loc_7112		; no, 16 or 32-bit
		and	ecx, 0FFh	; constrain to byte
		@RET

loc_7112:	test	byte ptr h_opflag+3[ebp], 40h	; 32 bit operand size?
		jnz	short loc_7121
		and	ecx, 0FFFFh	; constrain to word
		movsx	ecx, cx		; why here, but not on byte operand?
loc_7121:	@RET


; ===========================================================================
;		S u b r	o u t i	n e
;
; Write r/m operand
; E: ECX=operand value
;    h_modrm, h_disp
;    operand size in h_opflag[21], h_opflag[30]
;    segment override in h_opflag[26:24]
WriteRM		proc near
		movzx	ebx, byte ptr h_modrm[ebp]
		cmp	ebx, 0C0h-1
		jg	short WriteRMReg ; mod=3 (register)
		@CALL	WriteMem
		jmp	short WriteRMRet

WriteRMReg:	; mod=3 (register)
		and	ebx, 7		; R/M field
		@CALL	WriteReg
WriteRMRet:	@RET
WriteRM		endp


; ===========================================================================
;		S u b r	o u t i	n e
;
; Write value to register (shadow)
; E: ECX=operand value
;    EBX=register number
;    operand size in h_opflag[21], h_opflag[30]
; V: EAX, EBX
; R: byte/word/dword value in h_<reg>
WriteReg	proc near
		shl	ebx, 1		; *2 (size of short jump)
		test	byte ptr h_opflag+2[ebp], 20h	; byte operand?
		jnz	short WriteRegByte
		; word or dword operand
		add	ebx, offset WriteRegTableW
		@JMP	ebx

WriteRegByte:	; byte operand
		add	ebx, offset WriteRegTableB
		@JMP	ebx

WriteRegReturn:	@RET

WriteRegTableB:	; jumptable for write byte operand in CL to reg
		jmp	short WriteRegAL
		jmp	short WriteRegCL
		jmp	short WriteRegDL
		jmp	short WriteRegBL
		jmp	short WriteRegAH
		jmp	short WriteRegCH
		jmp	short WriteRegDH
		jmp	short WriteRegBH

WriteRegAL:	mov	byte ptr h_eax[ebp], cl
		jmp	short WriteRegReturn

WriteRegCL:	mov	byte ptr h_ecx[ebp], cl
		jmp	short WriteRegReturn

WriteRegDL:	mov	byte ptr h_edx[ebp], cl
		jmp	short WriteRegReturn

WriteRegBL:	mov	byte ptr h_ebx[ebp], cl
		jmp	short WriteRegReturn

WriteRegAH:	mov	byte ptr h_eax+1[ebp], cl
		jmp	short WriteRegReturn

WriteRegCH:	mov	byte ptr h_ecx+1[ebp], cl
		jmp	short WriteRegReturn

WriteRegDH:	mov	byte ptr h_edx+1[ebp], cl
		jmp	WriteRegRet2	; why not short WriteRegReturn here?

WriteRegBH:	mov	byte ptr h_ebx+1[ebp], cl
		jmp	short WriteRegReturn

WriteRegTableW:	; jumptable for write word/dword operand in CX/ECX to reg
		jmp	short WriteRegE_AX
		jmp	short WriteRegE_CX
		jmp	short WriteRegE_DX
		jmp	short WriteRegE_BX
		jmp	short WriteRegE_SPr
		jmp	short WriteRegE_BPr
		jmp	short WriteRegE_SIr
		jmp	short WriteRegE_DIr

; These are just redirections because these routines are out of short jump range.
WriteRegE_SPr:	jmp	WriteRegE_SP
WriteRegE_BPr:	jmp	WriteRegE_BP
WriteRegE_SIr:	jmp	WriteRegE_SI
WriteRegE_DIr:	jmp	WriteRegE_DI

WriteRegE_AX:	test	byte ptr h_opflag+3[ebp], 40h	; 32 bit operand size?
		jnz	short WriteRegEAX
		; Write AX
		mov	word ptr h_eax[ebp], cx
		jmp	WriteRegReturn

WriteRegEAX:	mov	h_eax[ebp], ecx
		jmp	WriteRegReturn

WriteRegE_CX:	test	byte ptr h_opflag+3[ebp], 40h	; 32 bit operand size?
		jnz	short WriteRegECX
		mov	word ptr h_ecx[ebp], cx
		jmp	WriteRegReturn

WriteRegECX:	mov	h_ecx[ebp], ecx
		jmp	WriteRegReturn

WriteRegE_DX:	test	byte ptr h_opflag+3[ebp], 40h	; 32 bit operand size?
		jnz	short WriteRegEDX
		mov	word ptr h_edx[ebp], cx
		jmp	WriteRegReturn

WriteRegEDX:	mov	h_edx[ebp], ecx
		jmp	WriteRegReturn

WriteRegE_BX:	test	byte ptr h_opflag+3[ebp], 40h	; 32 bit operand size?
		jnz	short WriteRegEBX
		mov	word ptr h_ebx[ebp], cx
		jmp	WriteRegReturn

WriteRegEBX:	mov	h_ebx[ebp], ecx
		jmp	WriteRegReturn

WriteRegE_SP:	test	byte ptr h_opflag+3[ebp], 40h	; 32 bit operand size?
		jnz	short WriteRegESP
		mov	word ptr h_esp[ebp], cx
		jmp	WriteRegReturn

WriteRegESP:	mov	h_esp[ebp], ecx
		jmp	WriteRegReturn

WriteRegE_BP:	test	byte ptr h_opflag+3[ebp], 40h	; 32 bit operand size?
		jnz	short WriteRegEBP
		mov	word ptr h_ebp[ebp], cx
		jmp	WriteRegReturn

WriteRegEBP:	mov	h_ebp[ebp], ecx
WriteRegRet2:	jmp	WriteRegReturn

WriteRegE_SI:	test	byte ptr h_opflag+3[ebp], 40h	; 32 bit operand size?
		jnz	short WriteRegESI
		mov	word ptr h_esi[ebp], cx
		jmp	short WriteRegRet2

WriteRegESI:	mov	h_esi[ebp], ecx
		jmp	short WriteRegRet2

WriteRegE_DI:	test	byte ptr h_opflag+3[ebp], 40h	; 32 bit operand size?
		jnz	short WriteRegEDI
		mov	word ptr h_edi[ebp], cx
		jmp	WriteRegReturn

WriteRegEDI:	mov	h_edi[ebp], ecx
		jmp	WriteRegReturn
WriteReg	endp


; ===========================================================================
;		S u b r	o u t i	n e
;
; Write memory operand
; E: ECX=operand value
;    h_disp
;    operand size in h_opflag[21], h_opflag[30]
;    segment override in h_opflag[26:24]
WriteMem	proc near
		xor	eax, eax
		mov	edx, h_disp[ebp]
		mov	al, byte ptr h_opflag+3[ebp]
		and	al, 7				; segment override
		shl	al, 1				; *2
		add	eax, h_base[ebp]
		test	byte ptr h_opflag+2[ebp], 20h	; byte operand?
		jnz	short WriteMemB
		test	byte ptr h_opflag+3[ebp], 40h	; 32 bit operand size?
		jnz	short WriteMemD
		add	eax, offset WriteMemTableW
		jmp	eax

WriteMemD:	; dword operand
		add	eax, offset WriteMemTableD
		jmp	eax

WriteMemB:	; byte operand
		add	eax, offset WriteMemTableB
		jmp	eax


WriteMemTableB:	; jumptable for byte operand
		jmp	short WriteMemBNone
		jmp	short WriteMemBES
		jmp	short WriteMemBCS
		jmp	short WriteMemBSS
		jmp	short WriteMemBDS
		jmp	short WriteMemBFS
		jmp	short WriteMemBGS
		jmp	short $			; fatal, should never be reached!

WriteMemBNone:
WriteMemBDS:	segds
		mov	[edx], cl
		jmp	short WriteMemBRet

WriteMemBCS:	mov	cs:[edx], cl
		jmp	short WriteMemBRet

WriteMemBSS:	mov	ss:[edx], cl
		jmp	short WriteMemBRet

WriteMemBES:	mov	es:[edx], cl
		jmp	short WriteMemBRet

WriteMemBFS:	mov	fs:[edx], cl
		jmp	short WriteMemBRet

WriteMemBGS:	mov	gs:[edx], cl
		jmp	short WriteMemBRet		; not needed

WriteMemBRet:	jmp	short WriteMemReturn


WriteMemTableW:	; jumptable for word operand
		jmp	short WriteMemWNone
		jmp	short WriteMemWES
		jmp	short WriteMemWCS
		jmp	short WriteMemWSS
		jmp	short WriteMemWDS
		jmp	short WriteMemWFS
		jmp	short WriteMemWGS
		jmp	short $			; fatal, should never be reached!

WriteMemWNone:
WriteMemWDS:	segds
		mov	[edx], cx
		jmp	short WriteMemWRet

WriteMemWCS:	mov	cs:[edx], cx
		jmp	short WriteMemWRet

WriteMemWSS:	mov	ss:[edx], cx
		jmp	short WriteMemWRet

WriteMemWES:	mov	es:[edx], cx
		jmp	short WriteMemWRet

WriteMemWFS:	mov	fs:[edx], cx
		jmp	short WriteMemWRet

WriteMemWGS:	mov	gs:[edx], cx
		jmp	short WriteMemWRet		; not needed

WriteMemWRet:	jmp	short WriteMemReturn


WriteMemTableD:	; jumptable for dword operand
		jmp	short WriteMemDNone
		jmp	short WriteMemDES
		jmp	short WriteMemDCS
		jmp	short WriteMemDSS
		jmp	short WriteMemDDS
		jmp	short WriteMemDFS
		jmp	short WriteMemDGS
		jmp	short $			; fatal, should never be reached!

WriteMemDNone:
WriteMemDDS:	segds
		mov	[edx], ecx
		jmp	short WriteMemReturn

WriteMemDCS:	mov	cs:[edx], ecx
		jmp	short WriteMemReturn

WriteMemDSS:	mov	ss:[edx], ecx
		jmp	short WriteMemReturn

WriteMemDES:	mov	es:[edx], ecx
		jmp	short WriteMemReturn

WriteMemDFS:	mov	fs:[edx], ecx
		jmp	short WriteMemReturn

WriteMemDGS:	mov	gs:[edx], ecx

WriteMemReturn:	@RET
WriteMem	endp


; ===========================================================================
;		S u b r	o u t i	n e
;
; Ensure check current privilege level is 0
; E:-
; V: EAX
; R: no return (raise exception general protection) if CPL not 0
CheckCPL0:	RDCPL	eax
		and	al, 3		; privilege level
		jnz	short RaiseGenProt	; <>0? -> exception general protection
		@RET

; raise exception general protection
RaiseGenProt:	push	EXCP_GENPROT+FORCEBIG
		push	0+FORCEBIG
		jmp	loc_2AA7

		ALIGN	256

hypseg		ends
		end

;-------------------------------------------------------------- end EMSUBS.ASM