;-----------------------------------------------------------------------------
; NexGen Nx586 BIOS Hypercode
;-----------------------------------------------------------------------------
; File: EXCPH.ASM - 
; Date: 2020-09-24
;
; Disassembled by Dipl.-Inf. Herbert Oppmann (herby@memotech.franken.de)
;-----------------------------------------------------------------------------

		include COMMON.INC

	extrn	FORCEBIG:abs
	extrn	loc_2046:near
	extrn	loc_2766:near

	public	ExcpDebugExcp
 IF VVL
	public	loc_2A20
 ENDIF
	public	loc_2A21
	public	loc_2AA7
	public	loc_2AC8
	public	loc_2B92
	public	loc_2BA4
	public	loc_2BB6
	public	loc_2BC5

 IF V033S
		db '@(#)excph.asm',TAB,'3.1',0
 ELSEIF V033V
		db '@(#)excph   ',0
 ELSE
		db '@(#)excph.asm',TAB,'3.2',0
 ENDIF
		ALIGN0	32

; E: [ESP].31-16 exception code
;    [ESP].15-0  selector
loc_2A20:	push	ebp

; Hypermode entry via offset 0800h
; E: [ESP]   saved EBP
;    [ESP+4].31-16 exception code
;    [ESP+4].15-0  selector
loc_2A21:	push	eax
		push	ebx
		push	ecx
		push	edx
		push	esi
		push	edi
		RDB	ebp, hbr	; = 0FFFF0800h
		add	ebp, 0C00h	; = HypStruc-HypEntries
 IF NOT V033S
		btr	word ptr h_dbflag[ebp], 2 ; is DR7 saved?
		jnc	short loc_2A41
		; restore DR7
		mov	eax, h_dr7[ebp]
		mov	dr7, eax
loc_2A41:
 ENDIF
		mov	eax, [esp+7*4]
		shr	eax, 16		; exception code
		cmp	al, 1		; #DB?
		jz	short loc_2A5E
		add	esp, 7*4	; drop saved registers
		xor	eax, eax
		OP_0FE4
		WRSBNK	eax
		; jump to exception handler routine
		mov	eax, h_base[ebp]
		add	eax, h_excph[ebp]
		jmp	eax

loc_2A5E:	; #DB debug exception
		bts	h_dbflag[ebp], 17	; generate #DB when exiting
		; restore registers
		pop	edi
		pop	esi
		pop	edx
		pop	ecx
		pop	ebx
		pop	eax
		pop	ebp
		; drop exception code
		add	esp, 4
		OP_0FE2

ExcpDebugExcp:	btr	h_dbflag[ebp], 17	; clear generate #DB
		btr	h_efl[ebp], 16		; clear RF resume flag
		btr	h_dbflag[ebp], 0	; clear #DB generated by single-stepping
		jnc	short loc_2A8C
		; #DB generated by single-stepping, i.e. set DR6.BS
		mov	eax, dr6
		or	ah, 40h
		mov	dr6, eax
loc_2A8C:	lea	esp, h_edi[ebp]
		pop	edi
		pop	esi
		pop	edx
		pop	ecx
		pop	ebx
		pop	eax
		pop	ebp
		add	esp, 4
		DELAYNOP
		INTH	EXCP_DEBUG


; E: [ESP].15-0   selector
;    [ESP+4].15-0 exception code
loc_2AA7:	pop	eax
		and	eax, 0FFFFh	; get selector
		pop	ebx		; get exception code
		shl	ebx, 16
		or	ebx, eax	; combine the two
		; build stack like exception occured by CPU
		push	0+FORCEBIG	; h_efl
		push	0+FORCEBIG	; h_eip3
		push	0+FORCEBIG	; h_gatflag
		push	ebx		; h_gate
		jmp	loc_2A20	; exception handler


; subroutine
loc_2AC8:	push	ebx
		push	ecx
		mov	ebx, offset loc_2766
		add	ebx, h_base[ebp]
		mov	eax, [esi+0Ch]
		cmp	eax, ebx
		jz	short loc_2AE8
		mov	ebx, offset loc_2046
		add	ebx, h_base[ebp]
		mov	eax, [esi+0Ch]
		cmp	eax, ebx
		jnz	short loc_2B05
loc_2AE8:	bts	h_dbflag[ebp], 1
		mov	eax, [esi]
		and	eax, 0FFFFFh
		cmp	eax, 0D0000h
		jnz	short loc_2B05
		mov	eax, 0A0000h
		jmp	loc_2B88

loc_2B05:	mov	eax, [esi]
		mov	ebx, eax
		shr	ebx, 16
		and	ebx, 0Fh
		bt	edx, 29
		jnc	short loc_2B41
		cmp	ebx, 0Dh
		jnz	short loc_2B88
		and	eax, 0FFF0FFFFh
		or	eax, 0A0000h
		bts	h_dbflag[ebp], 1
		mov	ebx, eax
		and	ebx, 0FFFFh
		jnz	short loc_2B88
		str	bx
		and	ebx, 0FFFCh
		mov	ax, bx
		jmp	short loc_2B88

loc_2B41:	cmp	ebx, 0Ch
		jnz	short loc_2B63
		RDCPL	ebx
		and	ebx, 3
		mov	cx, ss
		and	ecx, 3
		cmp	ebx, ecx
		jnz	short loc_2B88
		mov	ebx, ss
		and	ebx, 0FFFCh
		mov	ax, bx
		jmp	short loc_2B88

loc_2B63:	cmp	ebx, 0Dh
		jnz	short loc_2B88
		bt	edx, 30
		jnc	short loc_2B88
		RDCPL	ebx
		and	ebx, 3
		mov	cx, ss
		and	ecx, 3
		cmp	ebx, ecx
		jz	short loc_2B88
		and	eax, 0FFF0FFFFh
		or	eax, 0C0000h

loc_2B88:	mov	[esi], eax
		pop	ecx
		pop	ebx
		@RET


loc_2B92:	; #GP(selector)
		; general protection
		push	EXCP_GENPROT+FORCEBIG
		mov	eax, cs
		and	eax, 0FFFCh	; clear RPL
		push	eax
		jmp	loc_2AA7

loc_2BA4:	; #NP(selector)
		; segment not present
		push	EXCP_SEGNOTPRES+FORCEBIG
		mov	eax, cs
		and	eax, 0FFFCh	; clear RPL
		push	eax
		jmp	loc_2AA7

loc_2BB6:	; #GP(0)
		; general protection
		push	EXCP_GENPROT+FORCEBIG
		push	0+FORCEBIG
		jmp	loc_2AA7

; subroutine
; E: [ESP+4] = ?
; V: EAX, EBX, ECX, ESI
; R: no return if triple fault
;    [ESP+4] = 80000h on double fault
;    [ESP+4].0 set if ?
loc_2BC5:	mov	ecx, [esp+4]
		mov	eax, h_gatflag[ebp]
		bt	eax, 24		; set if error code
		jc	short loc_2BDD
		bt	eax, 25		; ??
		jc	short loc_2BEF
		jmp	short loc_2BF2	; do nothing

loc_2BDD:	; error code
		mov	ecx, [esp+4]
		mov	ebx, ecx
		@CALL	loc_2C0E
		jmp	short loc_2BF2

loc_2BEF:	; ??
		or	ecx, 1
loc_2BF2:	mov	[esp+4], ecx
		@RET


loc_2BFC:	; triple fault, shut down the CPU
		OP_0FE4
		xor	eax, eax
		WRAR	idtr, eax	; this will generate an exception
		DELAYNOP
		INTH	EXCP_DOUBLEFAULT

; subroutine
; E: EAX.20-16 = h_gatflag.exception/interrupt number (0-16)
;    EBX.19-16 = ?
;    ECX same as EBX
; V: EAX, EBX, ECX, ESI
; R: no return if triple fault
;    ECX = 80000h on double fault
;    ECX.0 set if ?
loc_2C0E:
		push	ecx			; save ECX
		shr	eax, 16
		and	eax, 1Fh		; exception/interrupt number
		shr	ebx, 16
		and	ebx, 0Fh
		; use exc/int number as index into table
		mov	esi, h_base[ebp]
		add	esi, offset dloc_2C4F
		mov	eax, [esi+eax*4]
		; use ?? as index into entry to address 2 bits
		mov	ecx, ebx
		add	ecx, ecx
		shr	eax, cl
		and	eax, 3
		pop	ecx			; restore ECX
		jz	short loc_2C49		; do nothing
		cmp	eax, 1
		jz	short loc_2C46		; set ECX.0
		cmp	eax, 2
		jz	short loc_2C3F		; double fault
		jmp	short near ptr loc_2BFC ; triple fault

loc_2C3F:	mov	ecx, 80000h
		jmp	short loc_2C49

loc_2C46:	or	ecx, 1
loc_2C49:	@RET


; for each exception/interrupt, there are 16 x 2 bits
; controls double/triple fault
; 00 = do nothing
; 01 = set ECX.0
; 02 = change to #DF double fault
; 03 = triple fault, shutdown CPU
dloc_2C4F:	;  0 #DE divide error
		dd 11001010101010110000000000000010b
		;  1 #DB debug
		dd 11000101010101110000000000000000b
		;  2     NMI interrupt
		dd 11000101010101110000000000000000b
		;  3 #BP breakpoint
		dd 11000101010101110000000000000000b
		;  4 #OF overflow
		dd 11000101010101110000000000000000b
		;  5 #BR BOUND range exceeded
		dd 11000101010101110000000000000000b
		;  6 #UD undefined opcode
		dd 11000101010101110000000000000000b
		;  7 #NM no math coprocessor
		dd 11000101010101110000000000000000b
		;  8 #DF double fault
		dd 11111111111111111111111111111111b
		;  9     coprocessor segment overrun
		dd 11001010101010110000000000000010b
		; 10 #TS invalid TSS
		dd 11001010101010110000000000000010b
		; 11 #NP segment not present
		dd 11001010101010110000000000000010b
		; 12 #SS stack segment fault
		dd 11001010101010110000000000000010b
		; 13 #GP general protection
		dd 11001010101010110000000000000010b
		; 14 #PF page fault
		dd 11101010101010110000000000000010b
		; 15     intel reserved, do not use
		dd 11111111111111111111111111111111b
		; 16 #MF math fault
		dd 11000101010101110000000000000000b

		ALIGN	256

hypseg		ends
		end

;--------------------------------------------------------------- end EXCPH.ASM