;-----------------------------------------------------------------------------
; NexGen Nx586 BIOS Hypercode
;-----------------------------------------------------------------------------
; File: FBLD.ASM - 
; Date: 2020-09-26
;
; Disassembled by Dipl.-Inf. Herbert Oppmann (herby@memotech.franken.de)
;-----------------------------------------------------------------------------

		include COMMON.INC

	extrn	Exception:near
	extrn	GetOperand:near
	extrn	loc_61DB:near
	extrn	loc_8ED1:near
	extrn	loc_8F53:near
	extrn	loc_90ED:near

	public	DecFBLD


 IF V033S OR V033V
		db '@(#)fbld.asm',TAB,'3.4',0
 ELSE
		db '@(#)fbld.asm',TAB,'3.6',0
 ENDIF
		ALIGN0	32

		dt 0FFFFC000000000000000h	; -QNaN

; Load Binary Coded Decimal
DecFBLD:	mov	h_excph[ebp], offset loc_8723
		or	byte ptr h_opflag+2[ebp], 10h	; set bit 20, only calculate effective address 
		@CALL	GetOperand
		xor	edx, edx	; unnecessary
		; read 10-byte BCD value into ebp-90h
		; -90h: D1,D0		-8Fh: D3,D2
		; -8Eh: D5,D4		-8Dh: D7,D6
		; -8Ch: D9,D8		-8Bh: D11,D10
		; -8Ah: D13,D12		-89h: D15,D14
		; -88h: D17,D16		-87h: sign, X
		mov	edx, h_disp[ebp]
		btr	h_opflag[ebp], 30	; set operand size to 16 bit
		btr	h_opflag[ebp], 21	; clear operand is byte
		add	edx, 8
		@CALL	loc_61DB	; read from EDX into BL/BX/EBX and h_content
		mov	[ebp-88h], bx	; byte nr. 8-9
		sub	edx, 8
		bts	h_opflag[ebp], 30	; set operand size to 32 bit
		@CALL	loc_61DB	; read from EDX into BL/BX/EBX and h_content
		mov	[ebp-90h], ebx	; byte nr. 0-3
		add	edx, 4
		@CALL	loc_61DB	; read from EDX into BL/BX/EBX and h_content
		mov	[ebp-8Ch], ebx	; byte nr. 4-7
		sub	esp, 9Eh
		mov	byte ptr [ebp-85h], 0	; sign = positive
		mov	word ptr [ebp-96h], 0	; buffer for FILD (integer -> float conversion)
		mov	dword ptr [ebp-9Ah], 10	; base of decimal system
		mov	eax, cr0
		mov	[ebp-4], eax
		or	eax, 1		; set PE = Protected mode enable
		mov	cr0, eax
		pushf
		btr	dword ptr [esp], 17	; clear VM = Virtual 8086 mode
		popf
		fnsave	byte ptr h_fenviron[ebp]
		FNLDSW	h_fsw[ebp]		; FPU Status Word
		FNLDTW	h_ftw[ebp]		; FPU Tag Word
		inc	h_fbldc[ebp]		; Statistics: Count number of FBLD instructions
		@CALL	loc_90ED
		cmp	cl, 1
 IF V033S OR V033V
		jz	loc_86F5
 ELSE
		jz	short loc_86F5
 ENDIF
		finit
		test	byte ptr [ebp-87h], 80h	; sign bit
		jz	short loc_86A1
		mov	byte ptr [ebp-85h], 1	; set sign negative

loc_86A1:
 IF V033S OR V033V
		; convert first byte
		mov	dword ptr [ebp-94h], 0	; power of 10
		movzx	eax, byte ptr [ebp-90h]	; read first byte
		and	al, 0Fh		; low digit
		mov	[ebp-96h], ax
		fild	word ptr [ebp-96h]
		inc	dword ptr [ebp-94h]
		mov	al, [ebp-90h]		; read first byte again
		and	al, 0F0h	; high digit
		sar	ax, 4
		mov	[ebp-96h], ax
		fild	word ptr [ebp-96h]
		fimul	dword ptr [ebp-9Ah]	; x 10
		faddp	st(1), st
		xor	ecx, ecx	; byte offset = 0
loc_9C5E:	; convert other bytes
		inc	dword ptr [ebp-94h]	; incr. power of 10
		inc	ecx			; incr. byte offset
		movzx	eax, byte ptr [ebp+ecx-90h] ; read byte
		mov	ebx, [ebp-94h]
		and	al, 0Fh		; low digit
		mov	[ebp-96h], ax
		fild	word ptr [ebp-96h]
loc_9C82:	fimul	dword ptr [ebp-9Ah]	; x 10
		dec	ebx
		cmp	ebx, 0
		ja	short loc_9C82
		faddp	st(1), st
		mov	al, [ebp+ecx-90h]	; read byte again
		and	al, 0F0h	; high digit
		sar	ax, 4
		mov	[ebp-96h], ax
		inc	dword ptr [ebp-94h]
		mov	ebx, [ebp-94h]
		fild	word ptr [ebp-96h]
loc_9CB6:	fimul	dword ptr [ebp-9Ah]	; x 10
		dec	ebx
		cmp	ebx, 0
		ja	short loc_9CB6
		faddp	st(1), st
		cmp	dword ptr [ebp-94h], 17	; done?
		jnc	short loc_86F5	; yes
		jmp	short loc_9C5E	; no, continue
 ELSE
		; same function, but more compact
		fldz
		fld1
		fild	word ptr [ebp-9Ah]	; = 10
		xor	ecx, ecx	; byte offset = 0
loc_86AD:	movzx	eax, byte ptr [ebp+ecx-90h] ; read byte
		and	al, 0Fh		; low digit
		mov	[ebp-96h], ax
		fild	word ptr [ebp-96h]
		fmul	st, st(2)
		faddp	st(3), st
		fmul	st(1), st
		movzx	eax, byte ptr [ebp+ecx-90h] ; read byte again
		and	al, 0F0h	; high digit
		sar	ax, 4
		mov	[ebp-96h], ax
		fild	word ptr [ebp-96h]
		fmul	st, st(2)
		faddp	st(3), st
		fmul	st(1), st
		inc	ecx
		cmp	ecx, 9		; done?
		jc	short loc_86AD
		fincstp
		fincstp
 ENDIF

loc_86F5:	fstp	tbyte ptr [ebp-90h]
		cmp	byte ptr [ebp-85h], 1	; negative?
		jnz	short loc_870B
		or	byte ptr [ebp-87h], 80h	; set sign
loc_870B:	@CALL	loc_8ED1
		fnrstor	byte ptr h_fenviron[ebp]
		fld	tbyte ptr [ebp-90h]	; load result
		jmp	loc_8F53

; Exception handler for FBLD
loc_8723:	jmp	Exception

		ALIGN0	16

hypseg		ends
		end

;---------------------------------------------------------------- end FBLD.ASM