;-----------------------------------------------------------------------------
; NexGen Nx586 BIOS Hypercode
;-----------------------------------------------------------------------------
; File: HYPRES.ASM - copy/uncompress hypercode into cache; reset processor
; Date: 2020-09-24
;
; Disassembled by Dipl.-Inf. Herbert Oppmann (herby@memotech.franken.de)
;-----------------------------------------------------------------------------

		include COMMON.INC

	extrn	FORCEBIG:abs

	extrn	loc_1820:near
	extrn	loc_1D20:near
	extrn	loc_1F20:near
	extrn	loc_2620:near
	extrn	loc_2A21:near
	extrn	loc_2D20:near
	extrn	loc_2D43:near
	extrn	loc_3720:near
	extrn	loc_3B20:near
	extrn	loc_3F20:near
	extrn	loc_6520:near
	extrn	loc_74E0:near
	extrn	EndOfCode:near

	public	SoftReset
	public	HypCodeVers
	public	loc_16D0

; subroutine call with ESP register
ECALL		macro dst:req
		local m
		mov	esp, offset m
		jmp	short dst
m:		endm

; return from subroutine called by ECALL
ERET		macro
		add	esp, M_BIOSBASE
		jmp	esp
		endm


; The Nx586 CPU starts in hypermode at address 0xFFFF0000. At this
; location the BIOS ROM contains a jmp instruction to a small routine
; which enables the lower 64KB of ROM and then jumps here, at M_BIOSBASE.

; First and second level caches are disabled when we come here from reset.

; Note:
; in initialisation routines, VL fetches values from hypvar structure via EBP,
; while PCI has these values in the code as immediate data. This is because
; the hypvar structure is still compressed in PCI and thus inaccessible.
;
; IF VVL
;		mov	edi, [ebx]
; ELSE	; PCI
;		mov	edi, 0FFFF0000h
; ENDIF

		; This code runs in hyper mode at address 0FFFE0000h
		; It copies/uncompresses itself to 2nd level cache so that
		; it appears in hyper mode at address 0FFFF0000h
		org	0000h

loc_0:	; do lot of self-test and initialisation
		clc
		jmp	short loc_14

SoftReset:	; skip self-test
		stc
		jmp	short loc_3B

		org	0010h		; db 10 dup(0)

; the following two values are calculated and inserted by an
; external program
; code checksum: the arithmetic sum of the uncompressed code words
; from 0000h to 07FFh and from 1800h to end of hypercode (EndOfCode)
; i.e. all hypercode except the hyperpage must be 0000h
		dw	0

; checksum magic: if == 0FFFFh, then above code checksum is not checked
chksmagic:	dw	0FFFFh		; magic 1234h (see flash program)
a_chksmagic	equ	offset chksmagic - offset loc_0

;-----------------------------------------------------------------------------

loc_14:		; clear registers
		DIAG	0A0h
		mov	eax,0
		mov	ecx,eax
		mov	edx,eax
		mov	ebx,eax
		mov	esp,eax
		mov	ebp,eax
		mov	esi,eax
		mov	edi,eax
	; v-- no effect --v
		mov	eax,ecx
		mov	ecx,eax
		mov	edx,eax
		mov	ebx,eax
		mov	esp,eax
		mov	ebp,eax
		mov	esi,eax
		mov	edi,eax
	; ^-- no effect --^

loc_3B:		DIAG	0A1h
		mov	eax,0
		WRSBNK	eax
		mov	eax,0
		WRCPL	eax
		; self-test wanted? (see code at loc_0)
		jnc	short loc_56	; yes, jump to test
		jmp	loc_437		; no, skip test

loc_56:		; self-test
		DIAG	0A2h
		mov	ebp, 1		; increment value

		DIAG	0A3h
		; test first level code/data cache tag area
		xor	eax, eax	; starting value =0
		mov	edx, 42000000h	; starting address
		mov	esi, CLINESIZE	; address increment
		mov	ecx, 80h	; loops
		ECALL	loc_BE
		neg	ebp		; decrement value
		dec	eax		; starting value = FFFFFFFF
		ECALL	loc_BE
		xor	ebp, ebp	; constant value
		ECALL	loc_BE

		DIAG	0A4h
		; test second level cache tag area and hypercode area
		mov	ebp, 1		; increment value
		mov	edx, 60C30000h	; starting address
		mov	esi, 4		; address increment
		mov	ecx, 4000h	; loops
		xor	eax, eax	; starting value
		ECALL	loc_BE
		neg	ebp		; decrement value
		dec	eax		; starting value = FFFFFFFF
		ECALL	loc_BE
		jmp	loc_14D

;-----------------------------------------------------------------------------

;		S u b r	o u t i	n e
;
; Test tag areas of cache memory
;
;E: ebp=value adjustment (1=increment, -1=decrement, 0=constant)
;   edx=starting address
;   esi=address increment
;   ecx=loops
;   eax=starting value
;V: EAX, BL, EDI, DR1, DR2, DR3
;R: EAX=0 if ok, no return on error

; if EDX=4200.0000h: test first level cache tag area
;   (every dword on 20h boundary) at
;   4200.0000h-4200.0FFFh, 4240.0000h-4240.0FFFh,
;   4280.0000h-4280.0FFFh, 42C0.0000h-42C0.0FFFh,
;   5200.0000h-5200.0FFFh, 5240.0000h-5240.0FFFh,
;   5280.0000h-5280.0FFFh, 52C0.0000h-52C0.0FFFh

; if EDX=60C3.0000h: test second level cache tag area and hypercode area
;   (bank 3 uppermost 64KB every dword) at
;   60C3.0000h-60C3.FFFFh

loc_BE		proc near
		; fill cache mem
		mov	dr2, ecx	; save loops
		mov	dr1, edx	; save starting address
		mov	dr3, eax	; save starting value
loc_C7:		mov	bl, 4
loc_C9:		mov	ecx, dr2	; restore loops
loc_CC:		mov	[edx], eax	; write
		add	edx, esi	; add address increment
		add	eax, ebp	; modify value
		loop	loc_CC
		xor	dx, dx
		cmp	edx, 60C30000h
		jge	short loc_FD
		add	edx, 400000h
		dec	bl
		jnz	short loc_C9
		add	edx, 10000000h
		and	edx, 0FF000000h
		cmp	edx, 60000000h
		jl	short loc_C7

loc_FD:		; check cache mem
		mov	eax, dr3	; restore starting value
		mov	edx, dr1	; restore starting address
loc_103:	mov	bl, 4
loc_105:	mov	ecx, dr2	; restore loops
loc_108:	mov	edi, [edx]
		cmp	edi, eax
		jnz	short $		; die on error!
		add	edx, esi	; add address increment
		add	eax, ebp	; modify value
		loop	loc_108
		xor	dx, dx
		cmp	edx, 60C30000h
		jge	short loc_13D
		add	edx, 400000h
		dec	bl
		jnz	short loc_105
		add	edx, 10000000h
		and	edx, 0FF000000h
		cmp	edx, 60000000h
		jl	short loc_103
loc_13D:	xor	eax, eax	; ok
		mov	ecx, dr2	; restore loops
		mov	edx, dr1	; restore starting address
		ERET
loc_BE		endp

;-----------------------------------------------------------------------------

loc_14D:	; detect size of second level cache
		DIAG	0A6h
		xor	eax, eax
		mov	bl, 1
		mov	edx, 60000000h
		mov	dword ptr [edx], 31415926h ; use PI as test pattern
		add	edx, 10000h
		cmp	dword ptr [edx], 31415926h ; repeats after 64K?
		jz	short loc_171	; yes, 256K cache memory
		mov	bl, 4		; no, 1M cache memory
		inc	eax
loc_171:	mov	dr0, eax	; save, 0=256K, 1=1M

		; clear second level cache tag area
		; 60C0C000 - 60C0FFFF (and 60C1C000 - 60C1FFFF,
		; 60C2C000 - 60C2FFFF and 60C3C000 - 60C3FFFF if 1M)
		; i.e. invalidate second level cache
		DIAG	0A7h
		mov	edi, 60C0C000h
		xor	eax, eax
loc_17F:	mov	ecx, 1000h
		cld			; up
	repe	stosd			; fill ECX dwords at [ES:EDI] with EAX
		or	edi, 0C000h
		dec	bl
		jnz	short loc_17F

		; copy/uncompress hypercode to second level cache,
		; bank 3, uppermost 64KB
		DIAG	0A8h
		mov	ebp, M_BIOSBASE + offset HypEntries
		add	ebp, HypStruc-HypEntries
		mov	edi, 60C00000h		; bank 3
 IF VVL
		; VL: everything is uncompressed
		lea	esi, h_base[ebp]
		mov	edx, [esi]
		and	edx, 3FFFFh
		or	edi, edx
		lea	ecx, [ebp+38h]		; unused in PCI
		mov	esi, [ecx]
		lea	ecx, h_endhyp[ebp]
		mov	ecx, [ecx]
		shr	ecx, 2
		cld
loc_0_1C3:	lodsd
		stosd
		loop	loc_0_1C3
 ELSE	; PCI
		; PCI: decompression code is uncompressed, following code is compressed
		mov	edx, 0FFFF0000h
		and	edx, 3FFFFh
		or	edi, edx	; edi = 60C30000h
		mov	esi, M_BIOSBASE
		cld
		mov	ecx, offset BeginCompr
		and	ecx, 0FFFFh
	repe	movsb
		lodsb
		dec	si		; compressed byte stream
		cmp	al, 7Fh		; must begin with Copy
		jz	short UCompLoop

		; fatal error, die
		DIAG	55h
		jmp	$

		; note: ECX=0 because of previous rep xxx
UCompLoop:	lodsb			; get control byte
		cmp	al, 0
		jnz	short UComp00
		; fill byte (small amount)
		lodsb			; get count
		mov	cl, al
UCompFillB:	lodsb
	repe	stosb
		jmp	short UCompLoop

UComp00:	cmp	al, 80h
		jnz	short UComp80
		; fill byte (large amount)
		lodsw
		mov	cx, ax
		jmp	short UCompFillB

UComp80:	cmp	al, 1
		jnz	short UComp01
		; fill word (small amount)
		lodsb
		mov	cl, al
UCompFillW:	lodsw
	repe	stosw
		jmp	short UCompLoop

UComp01:	cmp	al, 81h
		jnz	short UComp81
		; fill word (large amount)
		lodsw
		mov	cx, ax
		jmp	short UCompFillW

UComp81:	cmp	al, 7Fh
		jnz	short UComp7F
		; copy (small amount)
		lodsb
		mov	cl, al
UCompCopy: repe	movsb
		jmp	short UCompLoop

UComp7F:	cmp	al, 0FFh
		jnz	short UCompEnd		; typically by 0FEh
		; copy (large amount)
		lodsw
		mov	cx, ax
		jmp	short UCompCopy

UCompEnd:
 ENDIF
		; lock hypercode into second level cache by setting tags to
		; exclusive, make it appear at hypermode address 0FFFF0000h
		DIAG	0A9h
 IF VVL
		lea	ebx, h_base[ebp]
		mov	edi, [ebx]
 ELSE	; PCI
		mov	edi, 0FFFF0000h
 ENDIF
		mov	ebx, 0A000h	; pattern for tag
		;	1010.0000.0000.0000
		;	  ^------------------ A29* set here just to make sure
		;	                      this is for hypercode
		;	^^------------------- S1-S0 bits (MESI state: 10= exclusive)
 IF VVL
		lea	ecx, h_endhyp[ebp]
		mov	ecx, [ecx]
 ELSE	; PCI
		mov	ecx, offset EndOfCode
		and	ecx, 0FFFFh
 ENDIF
		add	ecx, edi	; end address

; now:
; EBX is tag pattern
; ECX is end address
; EDI is start address
loc_234:	mov	edx, 60C0C006h	; bit 2-0 = 6 = selects bank 3
		mov	esi, edi
		and	esi, 0C000h	; bit A15-A14
		shr	esi, 11		; -> bit 4-3 (C1-C0)
		or	edx, esi
		mov	esi, edi
		and	esi, 3FFE0h	; A17-A5
		or	edx, esi	; -> bit 17-5 (C14-C2)
		mov	esi, edi
		and	esi, 3FFF0000h	; A29-A16
		shr	esi, 16		; T13-T0
		or	ebx, esi	; really? Why does this work?
		mov	[edx], bx	; write tag
		add	edi, CLINESIZE
		cmp	ecx, edi
		ja	short loc_234

		; set configuration registers 0 and 1
		DIAG	0AAh
 IF VVL
		lea	ebx, h_cfg0[ebp]
		mov	ebx, [ebx]
		lea	ecx, h_cfg1[ebp]
		mov	ecx, [ecx]
 ELSE	; PCI
		mov	ebx, 5DF974C4h
		mov	ecx, 37021C0Fh
 ENDIF
		DELAYNOP 2
		WRCFG0	ebx
		WRCFG1	ecx
		OP_0FE4

		DIAG	0ABh
		mov	edx, 1C000000h
loc_296:	mov	eax, 7FFF0000h
		mov	[edx], eax
		or	edx, 1E000000h
		mov	eax, 0FFFF0000h
		mov	[edx], eax
		add	edx,   400000h
		and	edx, 1CC00000h
		cmp	edx, 1C000000h
		jnz	short loc_296
 IF VVL
		lea	edx, h_ccfg[ebp]	; cache configuration register
		mov	edx, [edx]
 ELSE	; PCI
		mov	edx, 8CD01300h
 ENDIF
		mov	eax, dr0		; second level cache memory size
		or	eax, eax		; 0=256K, 1=1M
		jz	short loc_2D0
		or	edx, BIT 28
loc_2D0:	; enable first and second level caches here
		mov	eax, 10000000h	; write cache configuration register
		mov	[eax], edx
		; Is this some special signal to the cache controller to start working?
		mov	eax, 0FFFF03F8h
	lock	add     byte ptr [eax], 0


		; From now on, the BIOS Flash-EPROM is not accessible from
		; hypercode anymore in the range 0FFFF0000h to end-of-hypercode!
		; Accesses end up in the second level cache.
		DIAG	0ACh
		cld			; not needed for PCI
 IF VVL
		mov	esi, ebp
		add	esi, 250h
		mov	ebx, 12000000h
		lodsd
		mov	[ebx], eax
		mov	ebx, 12400000h
		lodsd
		mov	[ebx], eax
		mov	esi, ebp
		add	esi, 25Ch
		mov	ebx, 12C00000h
		lodsd
		mov	[ebx], eax
 ELSE	; PCI
		mov	ebx, 12000000h
		mov	eax, 900AAh
		mov	[ebx], eax
		mov	ebx, 12400000h
		mov	eax, 1
		mov	[ebx], eax
		mov	ebx, 12C00000h
		mov	eax, 0
		mov	[ebx], eax
 ENDIF

		DIAG	0ADh
		cld
 IF VVL
		mov	ecx, 2
		mov	esi, ebp
		add	esi, 260h
		mov	ebx, 16000000h
loc_2CF:	lodsd
		mov	[ebx], eax
		add	ebx, 400000h
		loop	loc_2CF
		mov	ecx, 2
		mov	esi, ebp
		add	esi, 268h
		mov	ebx, 14000000h
loc_2EC:	lodsd
		mov	[ebx], eax
		add	ebx, 400000h
		loop	loc_2EC
		mov	ecx, 2
		mov	esi, ebp
		add	esi, 270h
		mov	ebx, 16800000h
loc_309:	lodsd
		mov	[ebx], eax
		add	ebx, 400000h
		loop	loc_309
		mov	ecx, 2
		mov	esi, ebp
		add	esi, 278h
		mov	ebx, 14800000h
loc_326:	lodsd
		mov	[ebx], eax
		add	ebx, 400000h
		loop	loc_326
		mov	esi, ebp
		add	esi, 258h
		mov	ebx, 12800000h
		lodsd
		mov	[ebx], eax
 ELSE	; PCI
		mov	ecx, 2
		xor	eax, eax
		mov	ebx, 16000000h
loc_31A:	mov	[ebx], eax
		add	ebx, 400000h
		loop	loc_31A
		mov	ecx, 2
		mov	ebx, 14000000h
loc_32E:	mov	[ebx], eax
		add	ebx, 400000h
		loop	loc_32E
		mov	ecx, 2
		mov	ebx, 16800000h
loc_342:	mov	[ebx], eax
		add	ebx, 400000h
		loop	loc_342
		mov	ecx, 2
		mov	ebx, 14800000h
loc_356:	mov	[ebx], eax
		add	ebx, 400000h
		loop	loc_356
		mov	eax, 0
		mov	ebx, 12800000h
		mov	[ebx], eax
 ENDIF

		; set HBR (hyper base register)
		DIAG	0AEh
 IF VVL
		lea	edi, [ebp+8]	; base address of segment register ?6 and ?7
					; after reset (= constant 0FFFF0800h)
		mov	eax, [edi]
 ELSE	; PCI
		mov	eax, 0FFFF0000h + offset HypEntries
 ENDIF
		mov	ebp, eax
		add	ebp, HypStruc-HypEntries
		WRB	hbr,eax

		; test TLB (translation lookaside buffer) (crippled)
		DIAG	0AFh
		mov	ebx, 0FFFFF010h
		;		    ^--- PL/REP: select TLB block 0
		;	      ^^^^^----- Physical address
		mov	eax, 000000340h
		;		     ^-- Command: write a TLB entry
		;		   ^^--- Flag bits D#,U,U# (Dirty, User)
		;	      ^^^^^----- Linear address
		mov	ecx, 4
loc_393:	mov	edx, ecx
		mov	ecx, 8
loc_39A:	mov	tr7, ebx
		mov	tr6, eax	; write TLB entry
		add	eax, 1000h	; increment linear address
		sub	ebx, 1000h	; decrement physical address
		loop	loc_39A
		add	ebx, 4		; next TLB block
		mov	ecx, edx
		loop	loc_393
		mov	ebx, 0		; unused
		mov	eax, 000000341h
		;		     ^-- Command: read a TLB entry
		;		   ^^--- Flag bits D#,U,U# (Dirty, User)
		;	      ^^^^^----- Linear address
		mov	ecx, 4*8
loc_3C3:	mov	tr6, eax
		mov	ebx, tr7
		mov	edx, tr6
		; TLB hit is actually not checked, so why the effort?
		add	eax, 1000h	; increment linear address
		loop	loc_3C3

		; check code checksum
		; (arithmetic sum of words over areas 0-7FFh and 1800h-end,
		; i.e. the whole hypercode (now in 2nd level cache!) except
		; the hyperpage)
		DIAG	0B0h
 IF VPCI
		mov	esi, h_base[ebp]
		cmp	word ptr a_chksmagic[esi], -1+FORCEBIG ; no checksum?
		jz	short loc_412	; then skip test
		mov	ecx, 400h
		xor	edx, edx
loc_3E9:	lodsw
		add	dx, ax
		loop	loc_3E9
		add	esi, 1000h
		lea	ecx, h_endhyp[ebp] ; end of hypercode
		mov	ecx, [ecx]
		sub	ecx, 1800h
		shr	ecx, 1
loc_406:	lodsw
		add	dx, ax
		loop	loc_406
		or	dx, dx
		jnz	$		; die on error

loc_412:
 ENDIF
		mov	eax, dr0	; second level cache memory size
		or	eax, eax	; 0=256K, 1=1M
		jz	short loc_437
		; 1M second level cache
		lea	esi, h_dbflag[ebp]
		or	dword ptr [esi], BIT 19
		lea	esi, h_ccfg[ebp]
		or	dword ptr [esi], BIT 28
		lea	esi, h_ccfgmaster[ebp]
		or	dword ptr [esi], BIT 28
		; end of self-test

loc_437:	; init all special registers
		DIAG	0B1h
		mov	eax, 7FFFFFF0h
		mov	cr0, eax
		mov	eax, 0
		mov	cr2, eax
		mov	cr3, eax
		mov	dr0, eax
		mov	dr1, eax
		mov	dr2, eax
		mov	dr3, eax
		mov	dr7, eax
		mov	dr6, eax
		mov	tr6, eax
		mov	tr7, eax

		; init all segment registers
		DIAG	0B2h
		xor	eax, eax
		WRSEL	es, eax
		lea	ebx, h_inisel[ebp]
		mov	ebx, [ebx]
		WRSEL	cs, ebx
		WRSEL	ss, eax
		WRSEL	ds, eax
		WRSEL	fs, eax
		WRSEL	gs, eax
		WRSEL	ldtr, eax
		WRSEL	tr, eax
		WRSEL	?6, eax
		WRSEL	?7, eax
		WRB     es, eax
		lea	ebx, h_inibase[ebp]
		mov	ebx, [ebx]
		WRB	cs, ebx
		WRB	ss, eax
		WRB	ds, eax
		WRB	fs, eax
		WRB	gs, eax
		WRB	ldtr, eax
		WRB	tr, eax
		WRB	gdtr, eax
		WRB	idtr, eax
		lea	eax, [ebp+8]
		mov	eax, [eax]
		WRB	?6, eax
		WRB	?7, eax
		mov	eax, 9200FFFFh
		mov	ebx, 9A00FFFFh
		WRAR	es, eax
		WRAR	cs, ebx
		WRAR	ss, eax
		WRAR	ds, eax
		WRAR	fs, eax
		WRAR	gs, eax
		WRAR	ldtr, eax
		WRAR	tr, eax
		WRAR	gdtr, eax
		mov	ecx, 920003FFh
		WRAR	idtr, ecx
		WRAR	?6, eax
		WRAR	?7, eax
 IF VPCI GT 'i'
		mov	dword ptr h_eflsh[ebp], 0
 ENDIF

		; detect math coprocessor presence
		DIAG	0B5h
		FWRCFG	h_fcfg[ebp]
		fnstsw	word ptr h_npx[ebp]	; get FPU status word
		mov	eax, h_npx[ebp]
		cmp	al, 10000001b
		;		   ^--- invalid operation
		;	    ^---------- error summary status
		jnz	short loc_558		; jump if FPU not available

		; init math coprocessor
		FWRCFG	h_fcfg[ebp]	; write configuration word
		FRDCFG	[ebp+5Ch]	; read it back
		add	h_fcfg[ebp], 40h
		FWRCFG	h_fcfg[ebp]
		FRDCFG	[ebp+60h]
		add	h_fcfg[ebp], 40h
		FWRCFG	h_fcfg[ebp]
		FRDCFG	[ebp+64h]
		add	h_fcfg[ebp], 40h
		FWRCFG	h_fcfg[ebp]
		FRDCFG	[ebp+68h]
 IF VPCI
		and	h_fcfg[ebp], 0FFFFFF3Fh ; undo the add 40h's
		FWRCFG	h_fcfg[ebp]
 ENDIF
		fninit
		mov	ecx, 4
loc_540:	fldz			; clear ST(0)-ST(7)
		fldz
		fldz
		fldz
		fldz
		fldz
		fldz
		fldz
		fninit
		loop	loc_540
 IF VVL OR V034I
		FNLDSW	h_npx[ebp]
 ENDIF
		OP_0FE4
 IF VPCI GT 'i'
		jmp	short loc_566
 ENDIF

loc_558:	; no math coprocessor
		clts
		mov	h_npx[ebp], 0DEADBEEFh
		DELAYIO
		mov	eax, h_npx[ebp]	; has no effect
 IF V033S
		LOPCMSK 7, 20h		; emulation for HLT
 ELSEIF V033V OR V033W
		LOPCMSK 6, 10h		; emulation for MOV CR,DR,TR
 ELSEIF V034I
 		; no change
 ELSE	; PCI
loc_566:	LOPCMSK	2, 0		; no emulation for POPF and PUSHF
 ENDIF

		; init general purpose registers and start mortal mode
		DIAG	0B7h
		RDMOST	eax		; get CPU model+stepping
		shr	eax, 24
		cmp	al, 0
		jnz	short loc_57F
		mov	eax, 3000000h
		mov	al, 3
loc_57F:	mov	ah, 15h
		mov	edx, eax	; CPU reset signature
		mov	eax, 0
		mov	ebx, eax
		mov	ecx, eax
		mov	esi, eax
		mov	edi, eax
		lea	esp, h_eip3[ebp]
		xor	ebp, ebp
		RETH

;-----------------------------------------------------------------------------

 IF VPCI
		; VL: everything is uncompressed
		db	'ASRE'		; magic for start of compressed stream
					; see end magic 'ASRD' in CPUSTUB.ASM
 ENDIF

BeginCompr:	db	'@(#) hcode sb.00',0
; That there are two identifying strings here could mean that a new
; source file is actually starting here.
; The names also suggest two things: '_hcode_s' (HyperCODE Startup?) for the
; processor startup in hyper mode and 'hypres' (HYPeR code EScapes) for entry
; from mortal mode into hyper mode.
; Even if there were two files, it doesn't look like there are two modules,
; because this location here is not aligned to a paragraph (16 bytes) or
; a page (256 byte) like the other modules!
; So maybe there are two source files, and one source file includes the other
; source file.

		db	'@(#)hypres.asm        b.03',0

; P05,768,575: A "block of DRAM memory for communication between hardware micro-
; code and software Hyper Code" is "used for all Hyper Mode entry points except
; Hard Reset"

; P05,905,997: "Alternately, this area could be used for storing branch pre-
; diction information, such as when a branch prediction cache in the micro-
; processor overflows."
; This could be another entry point (not enabled and not used here).

		org	0800h		; db 236h dup(0)
; HBR hyper base register points here, size is 4 KB
; Note: this can be moved
HypEntries:
; Exception
		push	ebp
		RDB	ebp, hbr	; = 0FFFF0800h
		add	ebp, 0C00h	; = HypStruc
		mov	ebp, h_base[ebp] ; = 0FFFF0000h
		add	ebp, offset loc_2A21	; EXCPH.ASM
		jmp	ebp

		org	0880h		; db 6Bh dup(0)
; Soft Reset
		push	ebp
		mov	ebp, offset loc_6520	; SOFRES.ASM
		add	ebp, h_base-h_ebp[esp]
		jmp	ebp

		org	0900h		; db 71h dup(0)
; Undefined opcodes
; P05,768,575: In order to gain control over the hardware breakpoint mechanisms
; and support both intermediate and physical address breakpoints, the F86 must
; be made aware of changes to the page tables. In order to do this, x86
; instructions which affect linear to physical mapping will be trapped as
; invalid opcodes. The original instruction can then be re-executed in Hyper
; Mode and appropriate information can be extracted for intermediate address
; breakpoint emulation.
		push	ebp
		mov	ebp, offset loc_3F20	; INVOP.ASM
		add	ebp, h_base-h_ebp[esp]
		jmp	ebp
; 0980h ??

; P05,768,575: Another possible reason for entering Hyper Mode is Protected
; Mode Support.
		org	0A00h		; db 0F1h dup(0)
; JMP and CALL instructions
		push	ebp
		mov	ebp, offset loc_2620	; CALL.ASM
		add	ebp, h_base-h_ebp[esp]
		jmp	ebp

		org	0A80h		; db 71h dup(0)
; Far Return
		push	ebp
		mov	ebp, offset loc_1820	; RETF.ASM
		add	ebp, h_base-h_ebp[esp]
		jmp	ebp

		org	0B00h		; db 71h dup(0)
; Interrupt
		push	ebp
		mov	ebp, offset loc_1F20	; INT.ASM
		add	ebp, h_base-h_ebp[esp]
		jmp	ebp

		org	0B80h		; db 71h dup(0)
; Interrupt Return
		push	ebp
		mov	ebp, offset loc_1D20	; IRET.ASM
		add	ebp, h_base-h_ebp[esp]
		jmp	ebp

		org	0C00h		; db 71h dup(0)
; Task switch
		push	ebp
		mov	ebp, offset loc_2D43	; TASK.ASM
		add	ebp, h_base-h_ebp[esp]
		jmp	ebp

		org	0C80h		; db 71h dup(0)
; I/O bit map
		push	ebp
		mov	ebp, offset loc_3720	; IOBM.ASM
		add	ebp, h_base-h_ebp[esp]
		jmp	ebp

		org	0D00h		; db 71h dup(0)
		push	ebp
		mov	ebp, offset loc_2D20	; TASK.ASM
		add	ebp, h_base-h_ebp[esp]
		jmp	ebp

		org	0D80h		; db 71h dup(0)
		push	ebp
		mov	ebp, offset loc_3B20	; INTO.ASM
		add	ebp, h_base-h_ebp[esp]
		jmp	ebp
; 0E00h ??
; 0E80h ??
		org	0F00h		; db 171h dup(0)
; P05,768,575: Another possible reason for entering Hyper Mode is to handle
; emulation of floating point instructions which are either not implemented
; in hardware, or whose implementations have not been thoroughly debugged.
		push	eax
		push	ebp
		mov	ebp, offset loc_74E0	; NPHCODE.ASM
		add	ebp, h_base-h_ebp[esp]
		jmp	ebp

		org	0F80h-4		; db 6Ch dup(0)
		db '@(#)'
HypCodeVers:
 IF V033V
		db '033s'		; this is a bug!
 ELSE
		db VERS			; e.g. '034q'
 ENDIF

		; this used as stack by SOFRES.ASM,
		; also for floating point operations, see e.g. F2XM1.ASM
	; hypvar structure
		org	1400h		; dd 287 dup(0)
HypStruc:	dd	0
		dd	0
		dd	0FFFF0000h + offset HypEntries
		dd	0FFFF0000h
; +10h
		dd	0
		dd	0
		dd	0
		dd	0
; +20h
		dd	0
		dd	0
		dd	0
		dd	0
; +30h
		dd	0
		dd	0
		dd	M_BIOSBASE
 IF VVL
		dd	84901000h	; copy of cache config register
 ELSE	; PCI
		dd	8CD01300h	; copy of cache config register
 ENDIF
; +40h
		dd	5DF974C4h	; copy of CFG0
 IF V033S
		dd	1507144Fh	; copy of CFG1
 ELSEIF V033V OR V033W
		dd	15071C4Fh	; copy of CFG1
 ELSE 	; PCI
		dd	37021C0Fh	; copy of CFG1
 ENDIF
		dd	0		; copy of opcode mask register
 IF V033V OR V033W
		dd	00100000h	;  "
 ELSE	; PCI+V033S
		dd	00000000h	;  "
 ENDIF
; +50h
 IF V033S
		dd	000F0000h	; segment base
 ELSE	; PCI+V033V+V033W
		dd	0FFFF0000h	; segment base
 ENDIF
		dd	0000F000h	; segment selector
		dd	0000FFF0h	; initial EIP
		dd	00800010h	; copy of floating point configuration register
; +60h
		dd	0
 IF VVL
		dd	3DE00000h
		dd	00280000h
 ELSE
		dd	00000000h
		dd	00000000h
 ENDIF
 IF VVL
		dd	00000006h	; FPU
 ELSEIF VPCI LT 'p'
		dd	00008006h	; FPU
 ELSE
		dd	00008004h	; FPU
 ENDIF
; +70h
		dd	0
		dd	0
		dd	0
		dd	0
; +80h
		dd	0
		dd	0
		dd	0
		dd	0
; +90h
		dd	0
		dd	0
		dd	0
		dd	0
; +A0h
		dd	0
		dd	offset EndOfCode	; end of uncompressed code
		dd	0
		dd	0
; +B0h
		dd	0
		dd	0
		dd	0
		dd	0
; +C0h
		dd	0
		dd	0
		dd	0
		dd	0
; +D0h
		dd	0
		dd	0
		dd	0
		dd	0
; +E0h
		dd	0
		dd	0
		dd	0
		dd	0
; +F0h
		dd	0
		dd	0
		dd	0
		dd	0
; +100h
		dd	0
		dd	0
		dd	0
		dd	0
; +110h
		dd	0
		dd	0
		dd	0
		dd	0
; +120h
		dd	0
		dd	0
		dd	0
		dd	0
; +130h
		dd	0
		dd	0
		dd	0
		dd	0
; +140h
		dd	0
		dd	0
		dd	0
		dd	0
; +150h
		dd	0
		dd	0
		dd	0
		dd	0
; +160h
		dd	0
		dd	0
		dd	0
		dd	0
; +170h
		dd	0
		dd	0
		dd	0
		dd	0
; +180h
		dd	0
		dd	0
		dd	0
		dd	0
; +190h
		dd	0
		dd	0
		dd	0
		dd	0
; +1A0h
		dd	0
		dd	0
		dd	0
		dd	0
; +1B0h
		dd	0
		dd	0
		dd	0
		dd	0
; +1C0h
		dd	0
		dd	0
		dd	0
		dd	0
; +1D0h
		dd	0
		dd	0
		dd	0
		dd	0
; +1E0h
		dd	0
		dd	0
		dd	0
		dd	0
; +1F0h
		dd	0
		dd	0
		dd	0
		dd	0
; +200h
		dd	0
		dd	0
		dd	0
		dd	0
; +210h
		dd	0
		dd	0
		dd	0
		dd	0
; +220h
		dd	0
		dd	0
		dd	0
		dd	0
; +230h
		dd	0
		dd	0
		dd	0
		dd	0
; +240h
		dd	0
		dd	0
		dd	0
		dd	0
; +250h
		dd	000900AAh
 IF V033V OR V033W
		dd	0
 ELSE	; PCI+V033S
		dd	1
 ENDIF
		dd	0
		dd	0
; +260h
		dd	0
		dd	0
		dd	0
		dd	0
; +270h
		dd	0
		dd	0
		dd	0
		dd	0
; +280h
 IF VVL
		dd	8ADF7000h
 ELSE	; PCI
		dd	8ADF7300h
 ENDIF
		dd	0		; copy of opcode mask register
		dd	0		;  "
		dd	0
; +290h
		dd	0
		dd	0
		dd	0
		dd	0
; +2A0h
		dd	0
		dd	0
		dd	0
		dd	0
; +2B0h
		dd	0
		dd	0
		dd	0
		dd	0
; +2C0h
		dd	0
		dd	0
		dd	0
		dd	0
; +2D0h
loc_16D0:	DELAYNOP 2
		LOPCMSK	 0, 0	; <-- modified by INVOP.ASM
		LOPCMSK	 1, 0
		LOPCMSK	 2, 0	; <-- modified by INVOP.ASM
		LOPCMSK	 3, 0	;     set to 6 for ID on (emulate PUSH/POP flags)
		LOPCMSK	 4, 0	;     and to 0 for ID off
		LOPCMSK	 5, 0
 IF V033V OR V033W
		LOPCMSK	 6, 10h
 ELSE	; PCI+V033S
		LOPCMSK	 6, 0
 ENDIF
		LOPCMSK	 7, 0
		LOPCMSK	 8, 0
		LOPCMSK	 9, 0
		LOPCMSK	10, 0
		LOPCMSK	11, 0
		LOPCMSK	12, 0
		LOPCMSK	13, 0
		LOPCMSK	14, 0
		LOPCMSK	15, 0
		@RET

		db	0, 0

		dd	0
		dd	0
; +330h
		dd	0
 IF V033V OR V033W
		dd	0FFFFFFFFh
 ELSE	; PCI+V033S
		dd	0
 ENDIF
		dd	0
		dd	0
; +340h
		dd	0
		dd	0
		dd	0
		dd	0
; +350h
		dd	0
		dd	0
		dd	0
		dd	0
; +360h
		dd	0
		dd	0
		dd	0
		dd	0
; +370h
		dd	0
		dd	0
		dd	0
		dd	0
; +380h
		dd	0
		dd	0
		dd	0
		dd	0
; +390h
		dd	0
		dd	0
		dd	0
		dd	0
; +3A0h
		dd	0
		dd	0
		dd	0
		dd	0
; +3B0h
		dd	0
		dd	0
		dd	0
		dd	0
; +3C0h
		dd	0
		dd	0
		dd	0
		dd	0
; +3D0h
		dd	0
		dd	0
		dd	0
		dd	0
; +3E0h
		dd	0
		dd	0
		dd	0
		dd	0
; +3F0h
		dd	0000FFF0h	; EIP
		dd	80000002h	; EFL
		dd	0		; ESP
		dd	0
; +400h

hypseg		ends
		end loc_0

;-------------------------------------------------------------- end HYPRES.ASM