;-----------------------------------------------------------------------------
; NexGen Nx586 BIOS Hypercode
;-----------------------------------------------------------------------------
; File: TASK.ASM - 
; Date: 2020-09-26
;
; Disassembled by Dipl.-Inf. Herbert Oppmann (herby@memotech.franken.de)
;-----------------------------------------------------------------------------

		include COMMON.INC

	extrn	FORCEBIG:abs

	extrn	loc_2AA7:near
	extrn	loc_2BB6:near
	extrn	loc_2BC5:near
	extrn	Exception:near
	extrn	ExcpDebugExcp:near
 IF VVL
	extrn	loc_40FD:near
 ENDIF

	public	loc_2D20
	public	loc_2D43
	public	loc_2D9A

; The task register holds the 16-bit segment selector, 32-bit base address,
; 16-bit segment limit, and descriptor attributes for the TSS of the current
; task. It references a TSS descriptor in the GDT.

 IF V033S
		db '@(#)task.asm',9,'3.5',0
 ELSEIF V033W
		db '@(#)task.asm',9,'3.6',0
 ELSE	; PCI+V033V
		db '@(#)task    ',0
 ENDIF
		ALIGN0	32

; Hypermode entry via offset 0D00h
loc_2D20:	mov	ebp, esp
		sub	ebp, 3E4h
 IF NOT V033S
		or	h_mode[ebp], 80000000h
 ENDIF
		push	eax
		push	ebx
		push	ecx
		push	edx
		push	esi
		push	edi
		mov	esp, ebp
		mov	dword ptr h_tsst[ebp], 08000000h
		jmp	short loc_2D9A

; Hypermode entry via offset 0C00h
loc_2D43:	mov	ebp, esp
		sub	ebp, 3E4h
 IF NOT V033S
		or	h_mode[ebp], 80000000h
 ENDIF
		push	eax
		push	ebx
		push	ecx
		push	edx
		push	esi
		push	edi
		mov	esp, ebp
		test	dword ptr h_gatflag[ebp], 60000000h
		jnz	short loc_2D82
		and	dword ptr h_gatflag[ebp], 7FFFFFFFh
		mov	eax, h_gatflag[ebp]
		mov	dword ptr h_tsst[ebp], 10000000h
		jmp	short loc_2D9A

loc_2D82:	and	dword ptr h_gatflag[ebp], 78000000h
		mov	eax, h_gatflag[ebp]
		and	eax, 60000000h
		mov	h_tsst[ebp], eax

loc_2D9A:	; h_tsst:
		; 30 \	set when entry via offset 0C00h
		; 29 /	from h_getflag.30-29 when not clear
		; 28	set when entry via offset 0C00h
		;	and h_gatflag.30-29 are clear
		; 27	set when entry via offset 0D00h
		mov	h_excph[ebp], offset loc_3543
		xor	edx, edx		; progress indicator
		mov	word ptr h_dbflag[ebp], 0
		@CALL	loc_3046
		bt	dword ptr h_tsst[ebp], 27
		jc	loc_2E3D

		; entry via offset 0C00h
		mov	eax, h_gatflag[ebp]
		and	eax, 18000000h	; bits 27-28
		shr	eax, 27		; old CS
		MOVSEL	ebx, ?7
		and	ebx, 0FFFCh	; 
		or	ebx, eax
		WRSEL	?7, ebx
		RDAR	eax, ?7
		and	eax, 1F000000h
		cmp	eax, 5000000h
		jnz	short near ptr loc_2DED
		RDB	eax, ?7
		jmp	short loc_2DEF

loc_2DED:	MOVSEL	eax, ?7		; CODE XREF: 2DE6

loc_2DEF:				; CODE XREF: 2DEB
		str	cx
		RDB	edi, tr
		RDAR	esi, tr
		ltr	ax
		OP_0FE4
		RDAR	eax, tr
		OP_0FE4
		WRAR	?7, eax
		RDB	eax, tr
		WRB	?7, eax
		RDSEL	eax, tr
		WRSEL	?7, eax
		WRSEL	tr, ecx
		WRB	tr, edi
		WRAR	tr, esi
		bts	word ptr h_dbflag[ebp], 1
		mov	ebx, 2Bh
		RDAR	eax, ?7
		bt	eax, 27
		jnc	short loc_2E33
		mov	ebx, 67h

loc_2E33:				; CODE XREF: 2E2C
		cmp	ax, bx
		jnc	short loc_2E3D
		jmp	near ptr loc_353F

loc_2E3D:	; entry via offset 0D00h
		bts	word ptr h_dbflag[ebp], 1
		xor	eax, eax
		OP_0FE4
		WRCPL	eax
		inc	eax
		WRSBNK	eax
		str	ax
		mov	h_trsel[ebp], eax
		MOVSEL	ecx, ?7
		mov	h_tsssel[ebp], ecx
		mov	esi, h_efl[ebp]
		bt	dword ptr h_tsst[ebp], 27
		jnc	short loc_2E6A
		btr	esi, 14

loc_2E6A:				; CODE XREF: 2E64
		bts	h_dbflag[ebp], 16
		jc	short loc_2E7B
		@CALL	loc_3100
loc_2E7B:	bt	dword ptr h_tsst[ebp], 27
		jnc	short loc_2E8C
		@CALL	loc_3496
loc_2E8C:	OP_0FE4
		RDAR	eax, ?7
		WRAR	tr, eax
		RDB	eax, ?7
		WRB	tr, eax
		xor	esi, esi
		test	dword ptr h_tsst[ebp], 30000000h
		jz	short near ptr loc_2EB4
		RDSEL	ebx, tr

loc_2EA8:				; DATA XREF: 361C
		mov	fs:0, bx
		mov	esi, 4000h
loc_2EB4:	OP_0FE4			; CODE XREF: 2EA3
		RDAR	eax, gs
		btr	eax, 31
		WRAR	gs, eax
		xor	eax, eax
		WRSEL	gs, eax
		RDAR	eax, fs
		btr	eax, 31
		WRAR	fs, eax
		xor	eax, eax
		WRSEL	fs, eax
		RDAR	eax, ds
		btr	eax, 31
		WRAR	ds, eax
		xor	eax, eax
		WRSEL	ds, eax
		RDAR	eax, ss
		btr	eax, 31
		WRAR	ss, eax
		xor	eax, eax
		WRSEL	ss, eax
		RDAR	eax, cs
		btr	eax, 31
		WRAR	cs, eax
		xor	eax, eax
		WRSEL	cs, eax
		RDAR	eax, es
		btr	eax, 31
		WRAR	es, eax
		xor	eax, eax
		WRSEL	es, eax
		@CALL	loc_3235
		mov	eax, cr0
		or	eax, 8
		mov	cr0, eax
		mov	eax, dr7
		and	eax, 0FFFFFEAAh
		mov	dr7, eax
		test	dword ptr h_tsst[ebp], 48000000h
		jz	short loc_2F41
		@CALL	loc_351B
loc_2F41:	OP_0FE4
		mov	eax, h_tsssel[ebp]
		WRSEL	tr, eax
		btr	h_dbflag[ebp], 16
		xor	eax, eax
		OP_0FE4
		WRSBNK	eax
		@CALL	loc_33CD
		bt	dword ptr h_tsst[ebp], 28
		jnc	short loc_2F70
		@CALL	loc_3071
loc_2F70:	bt	dword ptr h_tsst[ebp], 27
		jnc	short loc_2F7F
		bts	dword ptr h_efl[ebp], 31

loc_2F7F:				; CODE XREF: 2F75
		bt	dword ptr h_tsst[ebp], 31
		jc	loc_301B
		mov	ebx, h_eip3[ebp]
		RDAR	eax, cs
		bt	eax, 23
		jc	short loc_2FA0
		and	eax, 0FFFFFh
		jmp	short loc_2FA8

loc_2FA0:				; CODE XREF: 2F97
		shl	eax, 12
		or	eax, 0FFFh

loc_2FA8:				; CODE XREF: 2F9E
		cmp	ebx, eax
		ja	loc_2BB6
		btr	word ptr h_dbflag[ebp], 1
		push	ecx
		mov	ecx, dr7
 IF NOT V033S
		bts	word ptr h_dbflag[ebp], 2
		mov	h_dr7[ebp], ecx
 ENDIF
		push	ecx
		xor	ecx, ecx
		mov	dr7, ecx
		RDAR	ecx, cs
		bt	ecx, 25
		jc	short loc_2FED
		bts	ecx, 25
		OP_0FE4
		WRAR	cs, ecx
		test	byte ptr cs:[ebx], 0
		btr	ecx, 25
		OP_0FE4
		WRAR	cs, ecx
		jmp	short loc_2FF1

loc_2FED:				; CODE XREF: 2FD3
		test	byte ptr cs:[ebx], 0

loc_2FF1:				; CODE XREF: 2FEB
		pop	ecx
		mov	dr7, ecx
 IF NOT V033S
		btr	word ptr h_dbflag[ebp], 2
 ENDIF
		pop	ecx
		bt	h_dbflag[ebp], 17
		jc	ExcpDebugExcp
		lea	esp, h_edi[ebp]
		pop	edi
		pop	esi
		pop	edx
		pop	ecx
		pop	ebx
		pop	eax
		pop	ebp
		add	esp, 8
		OP_0FE4
		RETH

loc_301B:				; CODE XREF: 2F84
		mov	eax, dr6
		or	eax, 8000h
		and	eax, 0FFF0h
		mov	dr6, eax
		lea	esp, h_edi[ebp]
		pop	edi
		pop	esi
		pop	edx
		pop	ecx
		pop	ebx
		pop	eax
		pop	ebp
		add	esp, 4
		DELAYNOP
		INTH	EXCP_DEBUG

loc_3046:				; CODE XREF: 2DAE
		mov	eax, h_eip3[ebp]
		bt	dword ptr h_tsst[ebp], 28
		jnc	short loc_3065
		test	dword ptr h_gatflag[ebp], 3000000h
		jz	short loc_3065
		mov	eax, h_eip2[ebp]

loc_3065:				; CODE XREF: 3051	305D
		mov	h_eip1[ebp], eax
		@RET

loc_3071:				; CODE XREF: 2F6B
		mov	eax, h_gatflag[ebp]
		bt	eax, 24
		jnc	short loc_30E8
		mov	ebx, eax
		and	ebx, 0F0000h
		cmp	ebx, 80000h
		jl	short loc_30E8
		cmp	ebx, 0E0000h
		jg	short loc_30E8
		cmp	ebx, 90000h
		jz	short loc_30E8
		and	eax, 0FFFFh
		mov	esi, h_esp[ebp]
		RDAR	ebx, ss
		bt	ebx, 22
		jc	short near ptr loc_30B7
		and	esi, 0FFFFh
loc_30B7:	RDAR	ecx, tr		; CODE XREF: 30AF
		bt	ecx, 27
		jc	short loc_30C9
		sub	esi, 2
		mov	ss:[esi], ax
		jmp	short loc_30CF

loc_30C9:				; CODE XREF: 30BE
		sub	esi, 4
		mov	ss:[esi], eax

loc_30CF:				; CODE XREF: 30C7
		bt	ebx, 22
		jc	short loc_30E2
		mov	word ptr h_esp[ebp], si
		@RET

loc_30E2:				; CODE XREF: 30D3
		mov	h_esp[ebp], esi

loc_30E8:	@RET			; CODE XREF: 307B	308B
					; 00003093 0000309B

		ALIGN	256

; save registers into current TSS
; 32-bit: GS, FS, DS, SS, CS, ES,
;         EDI, ESI, EBP, ESP, EBX, EDX, ECX, EAX, EFLAGS, EIP
; 16-bit: DS, SS, CS, ES,
;         DI, SI, BP, SP, BX, DX, CX, AX, FLAGS, IP
; E: EFLAGS in ESI
loc_3100:	RDAR	eax, tr		; CODE XREF: 2E76
		bt	eax, 27		; 32 bit TSS?
		jnc	loc_31A4	; no, 16 bit

loc_310D:	; 32 bit TSS
		mov	fs:5Ch,	gs
		mov	fs:58h,	fs
		mov	fs:54h,	ds
		mov	fs:50h,	ss
		mov	fs:4Ch,	cs
		mov	fs:48h,	es
		mov	eax, h_edi[ebp]
		mov	fs:44h,	eax
		mov	eax, h_esi[ebp]
		mov	fs:40h,	eax
		mov	eax, h_ebp[ebp]
		mov	fs:3Ch,	eax
		mov	eax, h_esp[ebp]
		mov	fs:38h,	eax
		mov	eax, h_ebx[ebp]
		mov	fs:34h,	eax
		mov	eax, h_edx[ebp]
		mov	fs:30h,	eax
		mov	eax, h_ecx[ebp]
		mov	fs:2Ch,	eax
		mov	eax, h_eax[ebp]
		mov	fs:28h,	eax
		btr	esi, 31
		mov	fs:24h,	esi
		mov	eax, h_eip1[ebp]
		mov	fs:20h,	eax
		@RET

loc_31A4:	; 16 bit TSS
		mov	fs:28h,	ds
		mov	fs:26h,	ss
		mov	fs:24h,	cs
		mov	fs:22h,	es
		mov	eax, h_edi[ebp]
		mov	fs:20h,	ax
		mov	eax, h_esi[ebp]
		mov	fs:1Eh,	ax
		mov	eax, h_ebp[ebp]
		mov	fs:1Ch,	ax
		mov	eax, h_esp[ebp]
		mov	fs:1Ah,	ax
		mov	eax, h_ebx[ebp]
		mov	fs:18h,	ax
		mov	eax, h_edx[ebp]
		mov	fs:16h,	ax
		mov	eax, h_ecx[ebp]
		mov	fs:14h,	ax
		mov	eax, h_eax[ebp]
		mov	fs:12h,	ax
		mov	fs:10h,	si
		mov	eax, h_eip1[ebp]
		mov	fs:0Eh,	ax
		@RET


; restore registers from current TSS
; E: EFLAGS in ESI (is ORed to saved EFLAGS)
loc_3235:	RDAR	eax, tr		; CODE XREF: 2F15
		bt	eax, 27		; 32 bit TSS?
		jnc	loc_3313	; no, 16 bit

loc_3242:	; 32 bit TSS
		mov	eax, fs:60h
		WRSEL	ldtr, eax
		mov	eax, fs:5Ch
		WRSEL	gs, eax
		mov	eax, fs:58h
		WRSEL	fs, eax
		mov	eax, fs:54h
		WRSEL	ds, eax
		mov	eax, fs:50h
		WRSEL	ss, eax
		mov	eax, fs:4Ch
		WRSEL	cs, eax
		mov	eax, fs:48h
		WRSEL	es, eax
		mov	eax, fs:44h
		mov	h_edi[ebp], eax
		mov	eax, fs:40h
		mov	h_esi[ebp], eax
		mov	eax, fs:3Ch
		mov	h_ebp[ebp], eax
		mov	eax, fs:38h
		mov	h_esp[ebp], eax
		mov	eax, fs:34h
		mov	h_ebx[ebp], eax
		mov	eax, fs:30h
		mov	h_edx[ebp], eax
		mov	eax, fs:2Ch
		mov	h_ecx[ebp], eax
		mov	eax, fs:28h
		mov	h_eax[ebp], eax
		mov	eax, fs:24h
		or	eax, esi
		mov	h_efl[ebp], eax
		mov	eax, fs:20h
		mov	h_eip2[ebp], eax
		mov	h_eip3[ebp], eax
		mov	eax, fs:1Ch	; CR3
loc_32F5:	mov	ebx, fs:64h	; I/O map base address / T
		test	bl, 1		; debug trap flag T
		jz	short loc_3305
		bts	dword ptr h_tsst[ebp], 31
loc_3305:	mov	ebx, cr3
		cmp	eax, ebx	; no effect for PCI
 IF V033V OR V033W
		jz	short loc_320F
 ENDIF
		mov	cr3, eax
 IF VVL
loc_320F:
 ENDIF
		@RET

loc_3313:	; 16 bit TSS
		mov	ax, fs:2Ah
		WRSEL	ldtr, eax
		mov	ax, fs:28h
		WRSEL	ds, eax
		mov	ax, fs:26h
		WRSEL	ss, eax
		mov	ax, fs:24h
		WRSEL	cs, eax
		mov	ax, fs:22h
		WRSEL	es, eax
		mov	ax, fs:20h
		or	eax, 0FFFF0000h
		mov	h_edi[ebp], eax
		mov	ax, fs:1Eh
		mov	h_esi[ebp], eax
		mov	ax, fs:1Ch
		mov	h_ebp[ebp], eax
		mov	ax, fs:1Ah
		mov	h_esp[ebp], eax
		mov	ax, fs:18h
		mov	h_ebx[ebp], eax
		mov	ax, fs:16h
		mov	h_edx[ebp], eax
		mov	ax, fs:14h
		mov	h_ecx[ebp], eax
		mov	ax, fs:12h
		mov	h_eax[ebp], eax
		movzx	eax, word ptr fs:10h
		or	eax, esi
		mov	h_efl[ebp], eax
		movzx	eax, word ptr fs:0Eh
		mov	h_eip2[ebp], eax
		mov	h_eip3[ebp], eax
		@RET

loc_33CD:				; CODE XREF: 2F5A
		sldt	ax

loc_33D0:				; DATA XREF: 364C
		lldt	ax
		bt	dword ptr h_efl[ebp], 17
		jc	short loc_3426
		mov	ax, cs
		OP_0FE4
		WRCPL	eax
		mov	ax, ss

loc_33E8:				; DATA XREF: 3654
		mov	ss, ax
		OP_0FE4
		mov	ax, cs
loc_33EF:	MOVSEL	cs, eax		; DATA XREF: 365C
		OP_0FE4
		RDAR	eax, cs
		WRAR	cs, eax
		@CALL	loc_369C
		mov	ax, ds

loc_3406:				; DATA XREF: 3664
		mov	ds, ax
		mov	ax, es

loc_340B:				; DATA XREF: 366C
		mov	es, ax
		RDAR	eax, tr
		bt	eax, 27
		jnc	short loc_3420
		mov	ax, fs

loc_3419:				; DATA XREF: 3674
		mov	fs, ax
		mov	ax, gs

loc_341E:				; DATA XREF: 367C
		mov	gs, ax

loc_3420:	@RET			; CODE XREF: 3414

loc_3426:				; CODE XREF: 33DB
		mov	eax, 3
		OP_0FE4
		WRCPL	eax
		mov	eax, ss
		shl	eax, 4
		WRB	ss, eax
		mov	eax, 0F300FFFFh
		WRAR	ss, eax
		mov	eax, cs
		shl	eax, 4
		WRB	cs, eax
		mov	eax, 0FB00FFFFh
		WRAR	cs, eax
		mov	eax, ds
		shl	eax, 4
		WRB	ds, eax
		mov	eax, 0F300FFFFh
		WRAR	ds, eax
		mov	eax, es
		shl	eax, 4
		WRB	es, eax
		mov	eax, 0F300FFFFh
		WRAR	es, eax
		mov	eax, fs
		shl	eax, 4
		WRB	fs, eax
		mov	eax, 0F300FFFFh
		WRAR	fs, eax
		mov	eax, gs
		shl	eax, 4
		WRB	gs, eax
		mov	eax, 0F300FFFFh
		WRAR	gs, eax
		@RET

loc_3496:				; CODE XREF: 2E87
		mov	eax, fs:0
		mov	h_tsssel[ebp], eax
		bt	eax, 2
		jc	short loc_34FE
		test	eax, 0FFFCh
		jz	short loc_34FE
		mov	h_excph[ebp], offset loc_34B8
		OP_0FE4
loc_34B4:	MOVSEL	cs, eax		; DATA XREF: 3684
		jmp	short loc_34FE

loc_34B8:				; DATA XREF: 34AB
		mov	h_excph[ebp], offset loc_3543
		or	ebx, 1
		WRSBNK	ebx
		pop	ebx
		push	ebx
		shr	ebx, 16
		and	ebx, 0Fh
		cmp	ebx, 0Dh
		jnz	short loc_3543
		add	esp, 10h
		RDAR	eax, ?7
		mov	ebx, eax
		and	ebx, 1F000000h
		cmp	ebx, 0B000000h
		jnz	short loc_34EA
		jmp	short loc_34F2

loc_34EA:				; CODE XREF: 34E6
		cmp	ebx, 3000000h
		jnz	short loc_34FE

loc_34F2:				; CODE XREF: 34E8
		bt	eax, 31
		jnc	short loc_350A
		@RET

loc_34FE:				; CODE XREF: 34A2	34A9
					; 000034B6 000034F0
		mov	h_excph[ebp], offset loc_3543
		mov	eax, h_tsssel[ebp]
		jmp	short loc_3534

loc_350A:				; CODE XREF: 34F6
		push	EXCP_SEGNOTPRES
		MOVSEL	eax, ?7
		push	eax
		and	dword ptr [esp], 0FFFCh
		jmp	loc_2AA7

loc_351B:				; CODE XREF: 2F3C
		mov	eax, h_trsel[ebp]
		and	eax, 0FFF8h
		RDB	ebx, gdtr

loc_3526:				; DATA XREF: 368C
		btr	word ptr gs:[ebx+eax+4], 9
		@RET

loc_3534:				; CODE XREF: 3508	3541
					; 000036CF
		push	EXCP_INVALIDTSS+FORCEBIG
		push	eax
		jmp	loc_2AA7

loc_353F:	MOVSEL	eax, ?7		; CODE XREF: 2E38
		jmp	short loc_3534


; special exception handler for task
loc_3543:
 IF V033V OR V033W
		bts	h_dbflag[ebp], 22
		@CALL	loc_40FD
		mov	eax, h_cfg1[ebp]
		btr	eax, 18
		mov	h_cfg1[ebp], eax
		OP_0FE4
		WRCFG1	eax
 ENDIF
		mov	esi, esp
		@CALL	loc_35A4
		bt	h_dbflag[ebp], 16
		jnc	short loc_3581
		xor	eax, eax
		OP_0FE4
		WRCPL	eax
		inc	eax
		WRSBNK	eax
		mov	eax, [esi]
		shr	eax, 16
		and	eax, 0Fh
		RDB	ebx, idtr
		mov	ecx, gs:[ebx+eax*8+4]
		shr	ecx, 8
		and	cl, 1Fh
		cmp	cl, 5
		jz	short loc_3581
		btr	h_dbflag[ebp], 16

loc_3581:				; CODE XREF: 3554	357A
		mov	ax, cs
		OP_0FE4
		WRCPL	eax
		xor	eax, eax
		WRSBNK	eax
		bt	dword ptr h_tsst[ebp], 28
		jnc	short loc_359F
		@CALL	loc_2BC5
loc_359F:	jmp	Exception

loc_35A4:				; CODE XREF: 354A
		mov	ebx, [esi+0Ch]
		@CALL	loc_35F1
		cmp	dx, 0
		jz	short loc_35EB	; if not found
		mov	ebx, [esi]
		shr	ebx, 16
		and	ebx, 0Fh
		bt	edx, ebx
		jnc	short loc_35D0
		and	dword ptr [esi], 0FFF0FFFFh
		or	dword ptr [esi], 0A0000h
loc_35D0:	mov	eax, edx
		shr	eax, 16
		and	eax, 0FFh
		jz	short loc_35EB
		mov	ecx, h_tsssel[ebp]
		cmp	eax, 1
		jz	short loc_35EB
		mov	ecx, h_trsel[ebp]
		mov	[esi+4], cx
loc_35EB:	@RET


; subroutine: find fault(?) address in table
; E: EBX = fault address
; R: EDX = some bit-mapped code for this address (0= not found)
loc_35F1:	sub	ebx, h_base[ebp]
		mov	edi, h_base[ebp]
		add	edi, offset loc_361C
loc_35FD:	mov	eax, [edi]
		cmp	eax, ebx
		jz	short loc_3613
		add	edi, 8
		cmp	eax, 0FFFFFFFFh
		jnz	short loc_35FD
		xor	edx, edx
		@RET

loc_3613:	mov	edx, [edi+4]
		@RET

; table with addresses and bit-mapped codes
loc_361C	dd offset loc_2EA8
		dd 12000h
		dd offset loc_310D
		dd 22000h
		dd offset loc_31A4
		dd 22000h
		dd offset loc_3242
		dd 12000h
		dd offset loc_32F5
		dd 12000h
		dd offset loc_3313
		dd 12000h
		dd offset loc_33D0
		dd 2800h
		dd offset loc_33E8
		dd 2000h
		dd offset loc_33EF
		dd 2000h
		dd offset loc_3406
		dd 2000h
		dd offset loc_340B
		dd 2000h
		dd offset loc_3419
		dd 2000h
		dd offset loc_341E
		dd 2000h
		dd offset loc_34B4
		dd 12800h
		dd offset loc_3526
		dd 22000h
		dd 0FFFFFFFFh
		dd 0

loc_369C:	RDAR	ecx, cs		; CODE XREF: 33FE
		mov	esi, ecx
		shr	ecx, 29
		and	ecx, 3
		mov	ebx, cs
		and	ebx, 3
		bt	esi, 26
		jc	short loc_36B8
		sub	ecx, ebx
		jnz	short loc_36C8
		jmp	short loc_36BC

loc_36B8:				; CODE XREF: 36B0
		sub	ecx, ebx
		jg	short loc_36C8

loc_36BC:				; CODE XREF: 36B6
		bt	esi, 31
		jnc	short loc_36D4
		@RET

loc_36C8:				; CODE XREF: 36B4	36BA
		mov	eax, cs
		and	eax, 0FFFCh
		jmp	loc_3534

loc_36D4:				; CODE XREF: 36C0
		push	EXCP_SEGNOTPRES
		push	cs
		and	dword ptr [esp], 0FFFCh	; clear RPL
		jmp	loc_2AA7

		ALIGN	256

hypseg		ends
		end

;---------------------------------------------------------------- end TASK.ASM